/* FTP extension for TCP NAT alteration. */
#include <linux/module.h>
#include <linux/netfilter_ipv4.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <net/tcp.h>
#include "NAT/ip_nat.h"
#include "NAT/ip_nat_expect.h"
#include "NAT/ip_nat_helper.h"

EXPORT_NO_SYMBOLS;

/* Protects ftp info structure in conntrack */
static struct reserve_struct ftp_offset;

static DECLARE_LOCK(ftp_lock);

#ifndef HAVE_SKB_REALLOC_DATAROOM
#include <linux/config.h>
static void copy_skb_header(struct sk_buff *new, const struct sk_buff *old)
{
	/*
	 *	Shift between the two data areas in bytes
	 */
	unsigned long offset = new->data - old->data;

	new->list=NULL;
	new->sk=NULL;
	new->dev=old->dev;
	new->rx_dev=NULL;
	new->priority=old->priority;
	new->protocol=old->protocol;
	new->dst=dst_clone(old->dst);
	new->h.raw=old->h.raw+offset;
	new->nh.raw=old->nh.raw+offset;
	new->mac.raw=old->mac.raw+offset;
	memcpy(new->cb, old->cb, sizeof(old->cb));
	new->used=old->used;
	new->is_clone=0;
	atomic_set(&new->users, 1);
	new->pkt_type=old->pkt_type;
	new->stamp=old->stamp;
	new->destructor = NULL;
	new->security=old->security;
#ifdef CONFIG_NETFILTER
	new->nfmark=old->nfmark;
	new->nfreason=old->nfreason;
	new->nfcache=old->nfcache;
#ifdef CONFIG_NETFILTER_DEBUG
	new->nf_debug=old->nf_debug;
#endif
#endif
}

struct sk_buff *skb_realloc_dataroom(struct sk_buff *skb, int newdataroom)
{
	struct sk_buff *n;

	/*
	 *	Allocate the copy buffer
	 */
	
	n=alloc_skb((skb->data - skb->head) + newdataroom, GFP_ATOMIC);
	if(n==NULL)
		return NULL;

	/* Set the data pointer */
	skb_reserve(n,skb->data-skb->head);
	/* Set the tail pointer and length */
	skb_put(n,skb->len);
	/* Copy the bytes */
	memcpy(n->head,skb->head,skb->end-skb->head);
	copy_skb_header(n, skb);

	return n;
}
#endif

#if 0
#define DEBUGFTP printk
#else
#define DEBUGFTP(format, args...)
#endif

/* 75 seconds before we give up. */
#define FTP_TIMEOUT (75*HZ)

/* We keep track of where the last SYN correction was, and the SYN
   offsets before and after that correction.  Two of these (indexed by
   direction). */
struct ftp_info
{
	u_int32_t syn_correction_pos;
	int32_t syn_offset_before, syn_offset_after;
};

static const char *preamble[IP_CT_DIR_MAX]
= { "PORT ", "227 Entering Passive Mode (" };
static const char *postamble[IP_CT_DIR_MAX]
= { "\r\n", ")\r\n" };

/* This is interesting.  We simply use the port given us by the client
   or server.  In practice it's extremely unlikely to clash; if it
   does, the rule won't be able to get a unique tuple and will drop
   the packets. */
static int
mangle_packet(struct sk_buff **pskb,
	      const char *preamble,
	      u_int32_t newip,
	      struct ip_conntrack_expect *expected,
	      const char *postamble,
	      struct ftp_info *this_way,
	      struct ftp_info *other_way)
{
	struct iphdr *iph = (*pskb)->nh.iph;
	struct tcphdr *tcph = (void *)iph + iph->ihl*4;
	char *data = (void *)tcph + tcph->doff*4;
	size_t newlen, newtcplen;
	char buffer[25]; /* nnn,nnn,nnn,nnn,nnn,nnn */

	sprintf(buffer, "%u,%u,%u,%u,%u,%u",
		IP_PARTS(newip), ntohs(expected->tuple.dst.u.tcp.port)>>8,
		ntohs(expected->tuple.dst.u.tcp.port) & 0xFF);

	DEBUGFTP("Mapping tcph %p to new address: `%s'\n", tcph, buffer);
	newtcplen = strlen(preamble) + strlen(buffer) + strlen(postamble)
		+ tcph->doff*4;
	newlen = iph->ihl*4 + newtcplen;

	if (newlen > (*pskb)->len + skb_tailroom(*pskb)) {
		struct sk_buff *newskb;
		newskb = skb_realloc_dataroom(*pskb, newlen);
		if (!newskb) {
			DEBUGP("ftp: oom\n");
			return 0;
		} else {
			kfree_skb(*pskb);
			*pskb = newskb;
		}
	}

	/* SYN adjust.  If it's uninitialized, or this is after last
           correction, record it: we don't handle more than one
           adjustment in the window, but do deal with common case of a
           retransmit. */
	if (this_way->syn_offset_before == this_way->syn_offset_after
	    || before(this_way->syn_correction_pos, ntohl(tcph->seq))) {
		this_way->syn_correction_pos = ntohl(tcph->seq);
		this_way->syn_offset_before = this_way->syn_offset_after;
		this_way->syn_offset_after = (int32_t)
			this_way->syn_offset_before + newlen - (*pskb)->len;
	}

	/* Rewrite packet. */
	if (newlen > (*pskb)->len) {
		DEBUGFTP("ip_nat_ftp: Extending packet by %u to %u bytes\n",
			 newlen - (*pskb)->len, newlen);
		skb_put(*pskb, newlen - (*pskb)->len);
	} else {
		DEBUGFTP("ip_nat_ftp: Shrinking packet from %u to %u bytes\n",
			 (*pskb)->len, newlen);
		skb_trim(*pskb, newlen);
	}

	memcpy(data, preamble, strlen(preamble));
	memcpy(data + strlen(preamble), buffer, strlen(buffer));
	memcpy(data + strlen(preamble) + strlen(buffer), postamble,
	       strlen(postamble));
	iph->tot_len = htons(newlen);
	(*pskb)->csum = csum_partial((char *)tcph + tcph->doff*4,
				     newtcplen - tcph->doff*4, 0);
	tcph->check = 0;
	tcph->check = tcp_v4_check(tcph, newtcplen, iph->saddr, iph->daddr,
				   csum_partial((char *)tcph, tcph->doff*4,
						(*pskb)->csum));
	ip_send_check(iph);
	return 1;
}

/* Grrr... SACK.  Fuck me even harder.  Don't want to fix it on the
   fly, so blow it away. */
static void
delete_sack(struct sk_buff *skb, struct tcphdr *tcph)
{
	unsigned int i;
	u_int8_t *opt = (u_int8_t *)tcph;

	DEBUGFTP("Seeking SACKPERM in SYN packet (doff = %u).\n",
		 tcph->doff * 4);
	for (i = sizeof(struct tcphdr); i < tcph->doff * 4;) {
		DEBUGFTP("%u ", opt[i]);
		switch (opt[i]) {
		case TCPOPT_NOP:
		case TCPOPT_EOL:
			i++;
			break;

		case TCPOPT_SACK_PERM:
			goto found_opt;

		default:
			/* Worst that can happen: it will take us over. */
			i += opt[i+1] ?: 1;
		}
	}
	DEBUGFTP("\n");
	return;

 found_opt:
	DEBUGFTP("\n");
	DEBUGFTP("Found SACKPERM at offset %u.\n", i);

	/* Must be within TCP header, and valid SACK perm. */
	if (i + opt[i+1] <= tcph->doff*4 && opt[i+1] == 2) {
		/* Replace with NOPs. */
		tcph->check
			= ip_nat_cheat_check(*((u_int16_t *)(opt + i))^0xFFFF,
					     0, tcph->check);
		opt[i] = opt[i+1] = 0;
	}
	else DEBUGFTP("Something wrong with SACK_PERM.\n");
}

static inline int
handle_related(struct sk_buff **pskb,
	       struct ip_nat_info *info,
	       struct ip_conntrack_expect *expected,
	       struct ftp_info *ftp,
	       int dir)
{
	unsigned int i;
	u_int32_t newip = expected->tuple.dst.ip;
	struct ip_conntrack_tuple newtuple;
	/* This will rewrite it back to original */
	struct ip_conntrack_manip manip = { expected->tuple.dst.ip,
					    { expected->tuple.dst.u.tcp.port },
					    0 };

	for (i = 0; i < info->num_manips; i++) {
		/* We only worry about final source manipulations (we
		   could have both local-out and post-routing manips) */
		if (info->manips[i].direction == dir
		    && info->manips[i].maniptype == IP_NAT_MANIP_SRC)
			newip = info->manips[i].manip.ip;
	}

	if (!mangle_packet(pskb, preamble[dir], newip, expected,
			   postamble[dir], &ftp[dir], &ftp[!dir]))
		return 0;

	/* New tuple: destination IP will be different. */
	newtuple = expected->tuple;
	newtuple.dst.ip = newip;

	/* Tell NAT we need a fixed mapping for this expected conn. */
	if (!ip_nat_expect(&newtuple, &manip, IP_NAT_MANIP_DST, FTP_TIMEOUT)) {
		DEBUGP("ftp: ip_nat_expect failed!\n");
		return 0;
	}

	/* Now change conntrack's expectations. */
	expected->tuple.dst.ip = newip;
	return 1;
}

static unsigned int help(struct ip_conntrack_tuple_hash *h,
			 struct ip_nat_info *info,
			 enum ip_conntrack_info conntrackinfo,
			 unsigned int hooknum,
			 struct sk_buff **pskb)
{
	struct iphdr *iph = (*pskb)->nh.iph;
	struct tcphdr *tcph = (void *)iph + iph->ihl*4;
	u_int32_t newseq, newack;
	unsigned int ret;
	int dir = DIRECTION(h);
	struct ftp_info *ftp
		= STRUCT_RESERVED_PTR(ip_conntrack, h->ctrack, &ftp_offset);

	if (!ftp)
		return NF_ACCEPT;

	/* Delete SACK_OK on initial TCP SYNs. */
	if (tcph->syn && !tcph->ack)
		delete_sack(*pskb, tcph);

	LOCK_BH(&ftp_lock);
	/* conntrack code did the detection: we do data mangling in
           POST_ROUTING (since only source IP mangling matters). */
	if (conntrackinfo == IP_CT_ESTABLISHED_SPECIFIES_RELATED
	    && hooknum == NF_IP_POST_ROUTING) {
		if (!handle_related(pskb, info, &h->ctrack->expected, ftp,
				    dir)) {
			ret = NF_DROP;
			goto out;
		}
	}

	/* Only mangle packets once: original direction in POST_ROUTING
	   and reply direction on PRE_ROUTING. */
	if (!((hooknum == NF_IP_POST_ROUTING && dir == IP_CT_DIR_ORIGINAL)
	      || (hooknum == NF_IP_PRE_ROUTING && dir == IP_CT_DIR_REPLY))) {
		ret = NF_ACCEPT;
		goto out;
	}

	/* Sequence adjust */
	if (after(ntohl(tcph->seq), ftp[dir].syn_correction_pos))
		newseq = ntohl(tcph->seq) + ftp[dir].syn_offset_after;
	else
		newseq = ntohl(tcph->seq) + ftp[dir].syn_offset_before;
	newseq = htonl(newseq);

	/* Ack adjust */
	if (after(ntohl(tcph->ack_seq), ftp[!dir].syn_correction_pos))
		newack = ntohl(tcph->ack_seq) - ftp[!dir].syn_offset_after;
	else
		newack = ntohl(tcph->ack_seq) - ftp[!dir].syn_offset_before;
	newack = htonl(newack);

	tcph->check = ip_nat_cheat_check(~tcph->seq, newseq,
					 ip_nat_cheat_check(~tcph->ack_seq,
							    newack,
							    tcph->check));
	tcph->seq = newseq;
	tcph->ack_seq = newack;

	ret = NF_ACCEPT;
 out:
	UNLOCK_BH(&ftp_lock);
	return ret;
}

static struct ip_nat_helper ftp
= { { NULL, NULL }, IPPROTO_TCP, __constant_htons(21), help,
    NETFILTER_VERSION, "ftp" };

extern struct module *ip_conntrack_ftp;

int init_module(void)
{
	int ret;

	printk("ip_nat_ftp v%s\n", NETFILTER_VERSION);
	struct_reserve(ip_conntrack,
		       sizeof(struct ftp_info)*2,
		       __alignof__(struct ftp_info),
		       &ftp_offset);

	ret = ip_nat_helper_register(&ftp);
	if (ret == 0)
		__MOD_INC_USE_COUNT(ip_conntrack_ftp);
	return ret;
}

void cleanup_module(void)
{
	struct_unreserve(ip_conntrack, sizeof(struct ftp_info)*2, &ftp_offset);
	__MOD_DEC_USE_COUNT(ip_conntrack_ftp);
	ip_nat_helper_unregister(&ftp);
}
