/* Masquerade.  Simple mapping which alters range to a local IP address
   (depending on route). */
#include <linux/types.h>
#include <linux/ip.h>
#include <linux/timer.h>
#include <linux/module.h>
#include <linux/netfilter.h>
#include <net/protocol.h>
#include <net/checksum.h>
#include <linux/netfilter_ipv4.h>
#include "NAT/ip_nat_rule.h"

#define ASSERT_READ_LOCK(x) MUST_BE_READ_LOCKED(&ip_conntrack_lock)
#define ASSERT_WRITE_LOCK(x) MUST_BE_WRITE_LOCKED(&ip_conntrack_lock)

#include "common/listhelp.h"

EXPORT_NO_SYMBOLS;

/* Lock protects masq region inside conntrack */
static DECLARE_RWLOCK(masq_lock);

/* Offset in conntrack for masquerading. */
static struct reserve_struct masq_offset;

static int
masquerade_check(enum ip_nat_manip_type maniptype,
		 const struct ip_nat_rule *rule)
{
	if (maniptype != IP_NAT_MANIP_SRC) {
		DEBUGP("ipnat_rule_masquerade_check: must be SRC manip.\n");
		return 0;
	}

	return 1;
}

static unsigned int
masquerade_setup(struct ip_conntrack *conntrack,
		 struct ip_nat_info *info,
		 unsigned int hooknum,
		 struct sk_buff *skb,
		 struct ip_nat_multi_range *mr,
		 void *moptions)
{
	unsigned int ret;
	unsigned int i;
	u_int32_t newsrc;
	struct ip_nat_multi_range *newrange;
	int *masq_index
		= STRUCT_RESERVED_PTR(ip_conntrack, conntrack, &masq_offset);

	IP_CONNTRACK_ASSERT(hooknum == NF_IP_POST_ROUTING);
	if (!masq_index) {
		DEBUGP("Can't masquerade old conntrack %p\n", conntrack);
		return NF_DROP;
	}

	newsrc = inet_select_addr(skb->dst->dev,
				  ((struct rtable *)skb->dst)->rt_gateway,
				  RT_SCOPE_UNIVERSE);
	WRITE_LOCK(&masq_lock);
	*masq_index = skb->dst->dev->ifindex;
	WRITE_UNLOCK(&masq_lock);

	newrange = kmalloc(sizeof(struct ip_nat_multi_range)
			   + sizeof(struct ip_nat_range) * (mr->rangesize-1),
			   GFP_ATOMIC);
	if (!newrange) {
		DEBUGP("Masquerade: oom.\n");
		return NF_DROP;
	}

	/* Transfer from original range. */
	memcpy(newrange, mr,
	       sizeof(struct ip_nat_multi_range)
	       + sizeof(struct ip_nat_range) * (mr->rangesize-1));

	/* Substitute IPs for new IP dst. */
	for (i = 0; i < mr->rangesize; i++) {
		newrange->range[i].flags |= IP_NAT_RANGE_MAP_IPS;
		newrange->range[i].max_ip = newrange->range[i].min_ip
			= newsrc;
	}

	/* Hand modified range to generic setup. */
	ret = ip_nat_setup_info(conntrack, info, newrange, hooknum);
	kfree(newrange);
	return ret;
}

static inline int
device_cmp(const struct ip_conntrack_tuple_hash *h,
	   int ifindex)
{
	const int *masq_index;

	MUST_BE_READ_LOCKED(&masq_lock);
	if (h->ctrack->dead)
		return 0;

	masq_index = STRUCT_RESERVED_PTR(ip_conntrack, h->ctrack,&masq_offset);

	return masq_index && (*masq_index == ifindex);
}

static struct ip_conntrack_tuple_hash *
get_hash_by_ifindex(int ifindex)
{
	struct ip_conntrack_tuple_hash *h = NULL;
	size_t i;

	READ_LOCK(&masq_lock);
	READ_LOCK(&ip_conntrack_lock);
	for (i = 0; !h && i < IP_CONNTRACK_HTABLE_SIZE; i++) {
		h = LIST_FIND(&ip_conntrack_hash[i], device_cmp,
			      struct ip_conntrack_tuple_hash *, ifindex);
	}
	if (h)
		atomic_inc(&h->ctrack->use);
	READ_UNLOCK(&ip_conntrack_lock);
	READ_UNLOCK(&masq_lock);

	return h;
}

int masq_device_event(struct notifier_block *this,
		      unsigned long event,
		      void *ptr)
{
	struct net_device *dev = ptr;

	if (event == NETDEV_DOWN) {
		/* Device was downed.  Search entire table for
		   conntracks which were associated with that device,
		   and forget them. */
		struct ip_conntrack_tuple_hash *h;

		IP_CONNTRACK_ASSERT(dev->ifindex != 0);

		while ((h = get_hash_by_ifindex(dev->ifindex)) != NULL) {
			if (del_timer(&h->ctrack->timeout)) {
				IP_CONNTRACK_ASSERT(!h->ctrack->dead);
				h->ctrack->dead = 1;
				ip_conntrack_put(h->ctrack);
			}
			ip_conntrack_put(h->ctrack);
		}
	}

	return NOTIFY_DONE;
}

static struct notifier_block masq_dev_notifier = {
	masq_device_event,
	NULL,
	0
};


static struct ip_nat_mapping_type mapping_masquerade
= { { NULL, NULL }, "masquerade", NETFILTER_VERSION, THIS_MODULE,
    masquerade_check,
    masquerade_setup,
    0
};

int init_module(void)
{
	int ret;

	printk("ip_nat_map_masquerade v%s\n", NETFILTER_VERSION);
	ret = ip_nat_register_mapping_type(&mapping_masquerade);

	if (ret == 0) {
		struct_reserve(ip_conntrack, sizeof(int), __alignof__(int),
			       &masq_offset);
		/* Register for device down reports */
		register_netdevice_notifier(&masq_dev_notifier);
	}

	return ret;
}

void cleanup_module(void)
{
	ip_nat_unregister_mapping_type(&mapping_masquerade);
	struct_unreserve(ip_conntrack, sizeof(int), &masq_offset);
	unregister_netdevice_notifier(&masq_dev_notifier);
}
