/* Program to manage IP Nat rules.
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <net/if.h>
#include <netinet/in.h>
#include <getopt.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <stdarg.h>
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <linux/netfilter_ipv4.h>
#include <dlfcn.h>
#include "NAT/ip_nat.h"
#include "NAT/ip_nat_rule.h"
#include "NAT/ipnatctl.h"

#ifndef IPNAT_LIBDIR
#define IPNAT_LIBDIR "/usr/local/lib/ipnatctl"
#endif

#define NEW(type, contents...) 				\
({ 							\
	typeof (type *) _n = malloc(sizeof(type)); 	\
	if (_n) *_n = ( (type) ## contents);		\
	_n;						\
})

enum binding_type
{
	BINDING_UNSPECIFIED,
	BINDING_SOURCE,
	BINDING_DEST
};

static struct ipnat_proto *ipnat_protos = NULL;
static struct ipnat_mapping *ipnat_mappings = NULL;

static struct option original_opts[] = {
	{ "source", 1, 0, 's', },
	{ "destination", 1, 0, 'd' },
	{ "protocol", 1, 0, 'p' },
	{ "in-interface", 1, 0, 'i' },
	{ "out-interface", 1, 0, 'o' },
	{ "binding-type", 1, 0, 'b' },
	{ "to", 1, 0, 't' },
	{ "mapping-type", 1, 0, 'm' },
	{ "help", 0, 0, 'h' },
	{ "version", 0, 0, 'v' },
	{ "numeric", 0, 0, 'n' },
	{ NULL, 0, 0, 0 }
};

static struct option *opts = original_opts;

void print_usage(char *msg, ...)
{
	va_list args;

	va_start(args, msg);
	vfprintf(stderr, msg, args);
	va_end(args);

	fprintf(stderr,
		"\nipnatctl v%s\nUsage: "
		"   ipnatctl -[ID] [-n] <input spec> <mapping>\n"
		"   ipnatctl -L\n\n", NETFILTER_VERSION);

	fprintf(stderr,
		"<input spec>:\n"
		"  --source/-s address/mask		Source address\n"
		"  --destination/-d address/mask	Destination address\n"
		"  --in-interface/-i interface[+]       Incoming interface\n"
		"  --out-interface/-o interface[+]      Outgoing interface\n"
		"  --protocol/-p [name | number] [proto-specifics]\n"
		"					Protocol to match\n\n");
	fprintf(stderr,
		"<mapping>:\n"
		"  --binding-type/-b [source|destination]\n"
		"					Type of binding\n"
		"  --to/-t [address/mask | address-address]\n"
		"					IP address range to map to\n"
		"<mapping types>:\n"
		"  --mapping-type/-m <name>		What to do with matches.\n");

	exit(1);
}

/* FIXME */
int list_rules(int no_lookup)
{
	return system("cat /proc/net/ip_nat_rules");
}

static struct option *
merge_options(struct option *oldopts, struct option *newopts)
{
	unsigned int num_old, num_new;
	struct option *merge;

	for (num_old = 0; oldopts[num_old].name; num_old++);
	for (num_new = 0; newopts[num_new].name; num_new++);

	merge = malloc(sizeof(struct option) * (num_new + num_old + 1));
	memcpy(merge, oldopts, num_old * sizeof(struct option));
	memcpy(merge + num_old, newopts, (num_new) * sizeof(struct option));
	memset(merge + num_old + num_new, 0, sizeof(struct option));

	return merge;
}

static struct ipnat_proto *
find_protocol(int protocol, int tryload)
{
	struct ipnat_proto *ptr;

	for (ptr = ipnat_protos; ptr; ptr = ptr->next) {
		if (protocol == ptr->protonum)
			break;
	}

	if (!ptr && tryload) {
		char path[sizeof(IPNAT_LIBDIR) + sizeof("/libipnat_proto_.so")
			 + sizeof(int)*3];
		sprintf(path, IPNAT_LIBDIR "/libipnat_proto_%u.so", protocol);
		dlopen(path, RTLD_NOW);
		return find_protocol(protocol, 0);
	}
	return ptr;
}

static struct ipnat_mapping *
find_mapping(const char *name, int tryload)
{
	struct ipnat_mapping *ptr;

	for (ptr = ipnat_mappings; ptr; ptr = ptr->next) {
		if (strcmp(name, ptr->name) == 0)
			break;
	}

	if (!ptr && tryload) {
		char path[sizeof(IPNAT_LIBDIR) + sizeof("/libipnat_.so")
			 + strlen(name)];
		sprintf(path, IPNAT_LIBDIR "/libipnat_%s.so", name);
		dlopen(path, RTLD_NOW);
		return find_mapping(name, 0);
	}
	return ptr;
}

void
register_protocol(struct ipnat_proto *me)
{
	if (find_protocol(me->protonum, 0)) {
		fprintf(stderr, "ipnatctl: proto `%u' already registered.\n",
			me->protonum);
		exit(1);
	}

	/* Prepend to list. */
	me->next = ipnat_protos;
	ipnat_protos = me;

	opts = merge_options(opts, me->extra_opts);
}

void
register_mapping(struct ipnat_mapping *me)
{
	if (find_mapping(me->name, 0)) {
		fprintf(stderr, "ipnatctl: mapping `%s' already registered.\n",
			me->name);
		exit(1);
	}

	/* Prepend to list. */
	me->next = ipnat_mappings;
	ipnat_mappings = me;

	opts = merge_options(opts, me->extra_opts);
}

static int get_address(const char *string, u_int32_t *ip, u_int32_t *mask,
		       int no_lookup)
{
	struct in_addr addr;
	char *delim;

	delim = strchr(string, '/');
	if (delim) {
		struct in_addr addrmask;

		if (strchr(delim+1, '.')) {
			if (!inet_aton(delim+1, &addrmask))
				return 0;
			*mask = addrmask.s_addr;

			*delim = '\0';
			if (!inet_aton(string, &addr))
				return 0;
			*ip = addr.s_addr & *mask;

			return 1;
		} else {
			unsigned long tmpmask;

			errno = 0;
			tmpmask = strtoul(delim+1, NULL, 10);
			if (tmpmask == ULONG_MAX && errno)
				return 0;

			*mask = htonl((~0) << (32 - tmpmask));
			*delim = '\0';
			if (!inet_aton(string, &addr))
				return 0;

			*ip = addr.s_addr & *mask;

			return 1;
		}
	}
	else {
		if (!inet_aton(string, &addr)) {
			if (!no_lookup) {
				struct hostent *host;
				host = gethostbyname(string);
				
				if (host
				    && host->h_addrtype == AF_INET
				    && host->h_length
				    == sizeof(struct in_addr)) {
					if (host->h_addr_list[1] != NULL) {
						fprintf(stderr,
							"Host `%s' resolved to"
							" multiple addresses: NYI.\n",
							string);
						exit(1);
					}
					*ip = ((struct in_addr *)
					       host->h_addr_list[0])->s_addr;
					*mask = 0xFFFFFFFF;
					return 1;
				}
			}
			return 0;
		}
		*ip = addr.s_addr;
	}

	*mask = ~0;
	return 1;
}

static void
get_source_address(struct ip_nat_rule_user *rule, const char *optarg,
		   int no_lookup)
{
	if (!get_address(optarg, &rule->match.tuple.src.ip,
			 &rule->mask.tuple.src.ip, no_lookup))
		print_usage("Bad source address `%s'", optarg);
}

static void
get_destination_address(struct ip_nat_rule_user *rule,
			const char *optarg,
			int no_lookup)
{
	if (!get_address(optarg, &rule->match.tuple.dst.ip,
			 &rule->mask.tuple.dst.ip, no_lookup))
		print_usage("Bad destination address `%s'", optarg);
}

/* In case no /etc/protocols. */
struct pprot {
	char *name;
	unsigned short num;
};

static const struct pprot default_protos[] = {
	{ "all", 0 },
	{ "tcp", IPPROTO_TCP },
	{ "udp", IPPROTO_UDP },
	{ "icmp", IPPROTO_ICMP }
};

static void
get_protocol(struct ip_nat_rule_user *rule, const char *optarg, int no_lookup)
{
	int proto = atoi(optarg);

	if (proto < 1 || proto > 255) {
		struct protoent *pent;

		if (no_lookup)
			print_usage("illegal protocol `%s' specified",
				    optarg);

		pent = getprotobyname(optarg);
		if (pent)
			proto = pent->p_proto;
		else {
			unsigned int i;

			for (i = 0;
			     i < sizeof(default_protos)/sizeof(struct pprot);
			     i++) {
				if (strcasecmp(optarg, default_protos[i].name)
				    == 0)
					proto = default_protos[i].num;
			}

			if (i == sizeof(default_protos)/sizeof(struct pprot)) {
				print_usage("unknown protocol `%s' specified",
					    optarg);
			}
		}
	}

	rule->match.tuple.dst.protonum = proto;
	rule->mask.tuple.dst.protonum = 0xFF;

	/* This loads the protocol .so if available. */
	find_protocol(proto, 1);
}

static void
get_bind_type(struct ip_nat_rule_user *rule,
	      const char *optarg,
	      enum binding_type *btype)
{
	if (strncasecmp(optarg, "destination", strlen(optarg) ?: 1) == 0) {
		if (*btype == BINDING_SOURCE)
			print_usage("`-b %s' and `-o' incompatible", optarg);
		rule->maniptype = IP_NAT_MANIP_DST;
		*btype = BINDING_DEST;
	} else if (strncasecmp(optarg, "source", strlen(optarg) ?: 1) == 0) {
		if (*btype == BINDING_DEST)
			print_usage("`-b %s' and `-i' incompatible", optarg);
		rule->maniptype = IP_NAT_MANIP_SRC;
		*btype = BINDING_SOURCE;
	} else
		print_usage("Unknown binding type `%s'", optarg);
}

static int
get_address_range(const char *string, u_int32_t *ipmin, u_int32_t *ipmax)
{
	struct in_addr minaddr;
	char *delim;

	delim = strchr(string, '-');
	if (delim) {
		struct in_addr maxaddr;

		if (!inet_aton(delim+1, &maxaddr))
			return 0;
		*ipmax = maxaddr.s_addr;
		*delim = '\0';
	}
	else if ((delim = strchr(string, '/')) != NULL) {
		/* Here we figure our own interpretation of the pre-/
                   part of string so we can apply mask: hence
                   192.168.1.1/24 works as expected
                   (192.168.1.0-192.168.1.254). */
		if (strchr(delim+1, '.')) {
			struct in_addr addrmask;
			if (!inet_aton(delim+1, &addrmask))
				return 0;

			*delim = '\0';
			if (inet_aton(string, &minaddr) == 0)
				return 0;

			*ipmin = (minaddr.s_addr & addrmask.s_addr);
			/* Exclude broadcast address. */
			*ipmax = htonl(ntohl(minaddr.s_addr | ~addrmask.s_addr) - 1);
		} else {
			unsigned long mask;
			errno = 0;
			mask = strtoul(delim+1, NULL, 10);
			if (mask == ULONG_MAX && errno)
				return 0;

			mask = htonl((~0) << (32 - mask));
			*delim = '\0';
			if (inet_aton(string, &minaddr) == 0)
				return 0;

			*ipmin = (minaddr.s_addr & mask);
			/* Exclude broadcast address. */
			*ipmax = htonl(ntohl(minaddr.s_addr | ~mask) - 1);
		}
		return (ntohl(*ipmax) >= ntohl(*ipmin));
	}

	if (!inet_aton(string, &minaddr))
		return 0;
	*ipmin = minaddr.s_addr;
	if (!delim)
		*ipmax = *ipmin;

	return (ntohl(*ipmax) >= ntohl(*ipmin));
}

static void
get_to(struct ip_nat_rule_user *rule, const char *optarg)
{
	if (!get_address_range(optarg,
			       &rule->range.min_ip, &rule->range.max_ip))
		print_usage("Bad address range `%s'", optarg);
	rule->range.flags |= IP_NAT_RANGE_MAP_IPS;
}

static struct ip_nat_rule_user *
get_mapping(struct ip_nat_rule_user *rule, size_t *rulesize,
	    const char *optarg)
{
	if (strlen(optarg) >= IP_NAT_MAPPING_TYPE_MAX_NAMELEN)
		print_usage("Mapping `%s' too long", optarg);

	strcpy(rule->typename, optarg);
	find_mapping(rule->typename, 1);

	if (ipnat_mappings) {
		*rulesize = sizeof(*rule) + ipnat_mappings->optlen;
		rule = realloc(rule, *rulesize);
		if (!rule) {
			fprintf(stderr,
				"ipnatctl: Out of mem allocating %u bytes\n",
				(unsigned int)*rulesize);
			exit(1);
		}
	}
	return rule;
}

static void get_interface(char *ifname, char *ifmask, const char *name)
{
	size_t i;

	for (i = 0; i < IFNAMSIZ; i++) {
		switch (name[i]) {
		case '+':
			/* Wildcard. */
			memset(ifmask+i, 0, IFNAMSIZ-i);
			memset(ifname+i, 0, IFNAMSIZ-i);
			return;

		case 0:
			ifmask[i] = 0xFF;
			memset(ifmask+i+1, 0, IFNAMSIZ-i-1);
			memset(ifname+i, 0, IFNAMSIZ-i);
			return;

		default:
			ifname[i] = name[i];
			ifmask[i] = 0xFF;
		}
	}

	print_usage("Interface name `%s' must be less than %u chars\n",
		    name, IFNAMSIZ);
}

static struct ip_nat_rule_user *main_parse(int argc, char *argv[],
					   struct ip_nat_rule_user *rule,
					   size_t *rulesize,
					   int *op)
{
	int c;
	int no_lookup = 0;
	unsigned int mapping_flags = 0;
	enum binding_type btype = BINDING_UNSPECIFIED;
	int did_binding = 0;

	*op = 0;
	while ((c = getopt_long(argc, argv, "LIDns:d:p:m:t:b:i:o:", opts, NULL))
	       != -1) {
		switch (c) {
		case 'L':
			exit(list_rules(no_lookup));
			break;

		case 'I':
			*op = IP_NAT_SO_SET_INSERT;
			break;

		case 'D':
			*op = IP_NAT_SO_SET_DELETE;
			break;

		case 'n':
			no_lookup = 1;
			break;

		case 's':
			get_source_address(rule, optarg, no_lookup);
			break;

		case 'd':
			get_destination_address(rule, optarg, no_lookup);
			break;

		case 'p':
			get_protocol(rule, optarg, no_lookup);
			break;

		case 'i':
			if (btype == BINDING_SOURCE)
				print_usage("`-b source' and "
					    "`-i' incompatible");
			btype = BINDING_DEST;
			get_interface(rule->match.ifname,
				      rule->mask.ifname,
				      optarg);
			break;

		case 'o':
			if (btype == BINDING_DEST)
				print_usage("`-b destination' and "
					    "`-o' incompatible");
			btype = BINDING_SOURCE;
			get_interface(rule->match.ifname,
				      rule->mask.ifname,
				      optarg);
			break;

		case 'b':
			get_bind_type(rule, optarg, &btype);
			did_binding = 1;
			break;

		case 't':
			get_to(rule, optarg);
			break;

		case 'm':
			rule = get_mapping(rule, rulesize, optarg);
			break;

		case 'h':
			if (ipnat_protos && ipnat_protos->help)
				ipnat_protos->help();

			if (ipnat_mappings && ipnat_mappings->help)
				ipnat_mappings->help();

			print_usage("ipnatctl help\n");

		case 'v':
			printf("ipnatctl version %s", NETFILTER_VERSION);
			exit(0);

		default:
			if (ipnat_protos
			    && ipnat_protos->parse(c, argv, rule))
				break;

			if (ipnat_mappings
			    && ipnat_mappings->parse(c, argv, (void *)(rule+1),
						     &mapping_flags))
				break;

			print_usage("Unknown argument `%s'", optarg);
		}
	}

	if (ipnat_mappings)
		ipnat_mappings->finalize(rule, (void *)(rule+1),
					 mapping_flags);

	if (optind < argc)
		print_usage("unknown arguments found on commandline");

	if (!did_binding)
		print_usage("Need --binding-type/-b option.");

	if (!*op)
		print_usage("Need -L, -I or -D.");

	return rule;
}

int main(int argc, char *argv[])
{
	int sockfd, op;
	struct ip_nat_rule_user *rule;
	size_t rulesize = sizeof(*rule);

	rule = malloc(rulesize);
	memset(rule, 0, rulesize);

	if (argc < 2)
		print_usage("Too few options.");

	rule = main_parse(argc, argv, rule, &rulesize, &op);

#if 0
	if (rule->range.flags & IP_NAT_RANGE_MAP_IPS)
	fprintf(stderr, "Range for --to: %u.%u.%u.%u - %u.%u.%u.%u.\n",
		ntohl(rule->range.min_ip) >> 24,
		(ntohl(rule->range.min_ip) >> 16) & 0xFF,
		(ntohl(rule->range.min_ip) >> 8) & 0xFF,
		ntohl(rule->range.min_ip) & 0xFF,
		ntohl(rule->range.max_ip) >> 24,
		(ntohl(rule->range.max_ip) >> 16) & 0xFF,
		(ntohl(rule->range.max_ip) >> 8) & 0xFF,
		ntohl(rule->range.max_ip) & 0xFF);

	if (rule->range.flags & IP_NAT_RANGE_PROTO_SPECIFIED)
		fprintf(stderr, "Ports %u to %u.\n",
			ntohs(rule->range.min.tcp.port),
			ntohs(rule->range.max.tcp.port));
#endif

	sockfd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
	if (sockfd < 0) {
		perror("Opening raw socket");
		exit(1);
	}

	if (setsockopt(sockfd, IPPROTO_IP, op, rule, rulesize) < 0) {
		perror("setsockopt problem");
		return 1;
	}
	return 0;
}
