/* FTP extension for IP connection tracking. */
#define EXPORT_SYMTAB
#include <linux/module.h>
#include <linux/netfilter.h>
#include <linux/ip.h>
#include <net/checksum.h>
#include <net/tcp.h>

#define ASSERT_READ_LOCK(x) MUST_BE_READ_LOCKED(&ip_conntrack_lock)
#define ASSERT_WRITE_LOCK(x) MUST_BE_WRITE_LOCKED(&ip_conntrack_lock)

#include "conntrack/ip_conntrack_helper.h"
#include "common/listhelp.h"

#define SERVER_STRING "227 Entering Passive Mode ("
#define CLIENT_STRING "PORT "

#if 0
#define DEBUGFTP printk
#else
#define DEBUGFTP(format, args...)
#endif

#define IP_PARTS_NATIVE(n)			\
(unsigned int)((n)>>24)&0xFF,			\
(unsigned int)((n)>>16)&0xFF,			\
(unsigned int)((n)>>8)&0xFF,			\
(unsigned int)((n)&0xFF)

#define IP_PARTS(n) IP_PARTS_NATIVE(ntohl(n))

static struct {
	const char *pattern;
	size_t plen;
	char term;
} search[2] = {
	{ CLIENT_STRING, sizeof(CLIENT_STRING) - 1, '\r' },
	{ SERVER_STRING, sizeof(SERVER_STRING) - 1, ')' }
};

static int try_number(const char *data, size_t dlen, u_int32_t array[6],
		      char term)
{
	size_t i;

	/* Keep data pointing at next char. */
	for (i = 0; dlen; dlen--, data++) {
		if (*data >= '0' && *data <= '9') {
			array[i] = array[i]*10 + *data - '0';
		}
		else if (*data == ',')
			i++;
		else {
			/* Unexpected character; true if it's the
			   terminator and we're finished. */
			return (*data == term && i == 5);
		}
	}

	return 0;
}

/* Return 1 for match, 0 for accept, -1 for partial. */
static int find_pattern(const char *data, size_t dlen,
			const char *pattern, size_t plen,
			char term,
			u_int32_t array[6])
{
	if (dlen == 0)
		return 0;

	if (dlen < plen) {
		if (strnicmp(data, pattern, dlen) == 0)
			return -1;
		else return 0;
	}

	if (strnicmp(data, pattern, plen) != 0) {
#if 0
		size_t i;

		DEBUGFTP("ftp: string mismatch\n");
		for (i = 0; i < plen; i++) {
			DEBUGFTP("ftp:char %u `%c'(%u) vs `%c'(%u)\n",
				 i, data[i], data[i],
				 pattern[i], pattern[i]);
		}
#endif		
		return 0;
	}

	if (!try_number(data + plen, dlen - plen, array, term))
		return -1;

	return 1;
}

/* FIXME: This should be in userspace.  Later. */
static unsigned int help(const struct iphdr *iph, size_t len,
			 struct ip_conntrack_tuple_hash *h,
			 enum ip_conntrack_info *conntrackinfo)
{
	/* tcplen not negative guarenteed by ip_conntrack_tcp.c */
	struct tcphdr *tcph = (void *)iph + iph->ihl * 4;
	const char *data = (const char *)tcph + tcph->doff * 4;
	size_t tcplen = len - iph->ihl * 4;
	u_int32_t array[6] = { 0 };
	int dir = DIRECTION(h);

	/* Until there's been traffic both ways, don't look in packets. */
	if (*conntrackinfo != IP_CT_ESTABLISHED) {
		DEBUGFTP("ftp: Conntrackinfo = %u\n", *conntrackinfo);
		return NF_ACCEPT;
	}

	/* Not whole TCP header? */
	if (tcplen < sizeof(struct tcphdr) || tcplen < tcph->doff*4) {
		DEBUGFTP("ftp: tcplen = %u\n", (unsigned)tcplen);
		return NF_ACCEPT;
	}

	/* Checksum invalid?  Ignore. */
	/* FIXME: Source route IP option packets --RR */
	if (tcp_v4_check(tcph, tcplen, iph->saddr, iph->daddr,
			 csum_partial((char *)tcph, tcplen, 0))) {
		DEBUGFTP("ftp_help: bad csum: %p %u %u.%u.%u.%u %u.%u.%u.%u\n",
			 tcph, tcplen, IP_PARTS(iph->saddr),
			 IP_PARTS(iph->daddr));
		return NF_ACCEPT;
	}

	switch (find_pattern(data, tcplen - tcph->doff * 4,
			     search[dir].pattern,
			     search[dir].plen, search[dir].term, array)) {
	case -1: /* partial */
		/* We don't usually drop packets.  After all, this is
		   connection tracking, not packet filtering.
		   However, it is neccessary for accurate tracking in
		   this case. */
		DEBUGFTP("ip_conntrack_tcp_ftp: partial control packet.\n");
		return NF_DROP;

	case 0: /* no match */
		DEBUGFTP("ip_conntrack_ftp_help: no match\n");
		return NF_ACCEPT;
	}

	DEBUGFTP("ip_conntrack_ftp_help: match\n");
	/* match */
	h->ctrack->expected.tuple
		= ((struct ip_conntrack_tuple)
		   { { h->ctrack->tuplehash[!dir].tuple.src.ip,
		       { 0 }, 0 },
		     { htonl((array[0] << 24) | (array[1] << 16)
			     | (array[2] << 8) | array[3]),
		       { htons(array[4] << 8 | array[5]) },
		       IPPROTO_TCP }});

	/* We might already be in expect list; replace it. */
	WRITE_LOCK(&ip_conntrack_lock);
	if (list_inlist(&expect_list, &h->ctrack->expected))
		LIST_DELETE(&expect_list, &h->ctrack->expected);
	list_prepend(&expect_list, &h->ctrack->expected);
	WRITE_UNLOCK(&ip_conntrack_lock);
	*conntrackinfo = IP_CT_ESTABLISHED_SPECIFIES_RELATED;

	return NF_ACCEPT;
}

static struct ip_conntrack_helper ftp = { { NULL, NULL },
					  IPPROTO_TCP,
					  { __constant_htons(21) },
					  help };

int init_module(void)
{
	printk("ip_conntrack_ftp.o v%s\n", NETFILTER_VERSION);
	return ip_conntrack_helper_register(&ftp);
}

void cleanup_module(void)
{
	ip_conntrack_helper_unregister(&ftp);
}

struct module *ip_conntrack_ftp = THIS_MODULE;
EXPORT_SYMBOL(ip_conntrack_ftp);

