/* Connection state tracking for netfilter.  This is separated from,
   but required by, the NAT layer; it can also be used by an iptables
   extension. */

/* (c) 1999 Paul `Rusty' Russell.  Licenced under the GNU General
   Public Licence. */

#define __NO_VERSION__
#include <linux/config.h>
#include <linux/types.h>
#include <linux/ip.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <net/checksum.h>
#include <linux/stddef.h>

#define ASSERT_READ_LOCK(x) MUST_BE_READ_LOCKED(&ip_conntrack_lock)
#define ASSERT_WRITE_LOCK(x) MUST_BE_WRITE_LOCKED(&ip_conntrack_lock)

#include "conntrack/struct_reserve.h"
#include "conntrack/ip_conntrack.h"
#include "conntrack/ip_conntrack_protocol.h"
#include "conntrack/ip_conntrack_helper.h"
#include "conntrack/ip_conntrack_core.h"
#include "common/listhelp.h"

DECLARE_RWLOCK(ip_conntrack_lock);

#ifdef DEBUG_CONNTRACK
static void dump_tuple(const struct ip_conntrack_tuple *tuple) __attribute__((unused));

static void dump_tuple(const struct ip_conntrack_tuple *tuple)
{
	printk("tuple %p: %u %u.%u.%u.%u:%u -> %u.%u.%u.%u:%u\n",
	       tuple, tuple->dst.protonum,
	       IP_PARTS(tuple->src.ip), ntohs(tuple->src.u.tcp.port),
	       IP_PARTS(tuple->dst.ip), ntohs(tuple->dst.u.tcp.port));
}
#endif

void (*ip_conntrack_destroyed)(struct ip_conntrack *conntrack) = NULL;
LIST_HEAD(expect_list);
LIST_HEAD(protocol_list);
static LIST_HEAD(helpers);
struct list_head ip_conntrack_hash[IP_CONNTRACK_HTABLE_SIZE];
static unsigned int ip_conntrack_gen = NF_REASON_MIN_RESERVED_FOR_CONNTRACK+1;

extern struct ip_conntrack_protocol ip_conntrack_generic_protocol;

DECLARE_STRUCT_RESERVABLE(ip_conntrack);

static inline int proto_cmpfn(const struct ip_conntrack_protocol *curr,
			      u_int8_t protocol)
{
	return protocol == curr->proto;
}

struct ip_conntrack_protocol *__find_proto(u_int8_t protocol)
{
	struct ip_conntrack_protocol *p;

	MUST_BE_READ_LOCKED(&ip_conntrack_lock);
	p = LIST_FIND(&protocol_list, proto_cmpfn,
		      struct ip_conntrack_protocol *, protocol);
	if (!p)
		p = &ip_conntrack_generic_protocol;

	return p;
}

struct ip_conntrack_protocol *find_proto(u_int8_t protocol)
{
	struct ip_conntrack_protocol *p;

	READ_LOCK(&ip_conntrack_lock);
	p = __find_proto(protocol);
	READ_UNLOCK(&ip_conntrack_lock);
	return p;
}

static inline u_int32_t
hash_conntrack(const struct ip_conntrack_tuple *tuple)
{
#if 0
	dump_tuple(tuple);
#endif
#ifdef DEBUG_CONNTRACK
	if (tuple->src.pad)
		DEBUGP("Tuple %p has non-zero padding.\n", tuple);
#endif
	/* ntohl because more differences in low bits. */
	/* To ensure that halves of the same connection don't hash
	   clash, we add the source per-proto again. */
	return (ntohl(tuple->src.ip + tuple->dst.ip
		     + tuple->src.u.all + tuple->dst.u.all
		     + tuple->dst.protonum)
		+ ntohs(tuple->src.u.all))
		% IP_CONNTRACK_HTABLE_SIZE;
}

inline int
get_tuple(const struct iphdr *iph, size_t len,
	  struct ip_conntrack_tuple *tuple,
	  struct ip_conntrack_protocol *protocol)
{
	int ret;

	/* FIXME:  Only need this check for icmps.  --RR */
	if (ntohs(iph->frag_off) & IP_OFFSET) {
		static int warned = 0;
		if (!warned)
			printk("ip_conntrack_core: Frag of proto %u.\n",
			       iph->protocol);
		warned = 1;
		return 0;
	}
	/* Guarentee 8 protocol bytes: if more wanted, use len param */
	else if (iph->ihl * 4 + 8 > len)
		return 0;

	tuple->src.ip = iph->saddr;
	tuple->src.pad = 0;
	tuple->dst.ip = iph->daddr;
	tuple->dst.protonum = iph->protocol;

	ret = protocol->pkt_to_tuple((u_int32_t *)iph + iph->ihl,
				     len - 4*iph->ihl,
				     tuple);
	return ret;
}

static inline int
count(const struct ip_conntrack_tuple_hash *i, unsigned int *c)
{
	(*c)++;
	return 0;
}

static inline int
conntrack_cmp(const struct ip_conntrack_tuple_hash *i,
	      const struct ip_conntrack_tuple *tuple)
{
	return !i->ctrack->dead
		&& (memcmp(tuple, &i->tuple, sizeof(*tuple)) == 0);
}

/* find a connection corresponding to a tuple. */
static struct ip_conntrack_tuple_hash *
find_conntrack_get(const struct ip_conntrack_tuple *tuple)
{
	struct ip_conntrack_tuple_hash *h;
	u_int32_t hash = hash_conntrack(tuple);

	READ_LOCK(&ip_conntrack_lock);
	h = LIST_FIND(&ip_conntrack_hash[hash],
		      conntrack_cmp, struct ip_conntrack_tuple_hash *, tuple);
	if (h)
		atomic_inc(&h->ctrack->use);
	READ_UNLOCK(&ip_conntrack_lock);

	return h;
}

static int
invert_tuple(struct ip_conntrack_tuple *inverse,
	     const struct ip_conntrack_tuple *orig,
	     const struct ip_conntrack_protocol *protocol)
{
	inverse->src.ip = orig->dst.ip;
	inverse->src.pad = 0;
	inverse->dst.ip = orig->src.ip;
	inverse->dst.protonum = orig->dst.protonum;

	return protocol->invert_tuple(inverse, orig);
}

void ip_conntrack_put(struct ip_conntrack *ct)
{
#ifdef DEBUG_CONNTRACK
	if (atomic_read(&ct->use) == 0)
		DEBUGP("Conntrack %p use goes negative.\n", ct);
#endif
	if (atomic_dec_and_test(&ct->use)) {
		IP_CONNTRACK_ASSERT(ct->dead);
		IP_CONNTRACK_ASSERT(!timer_pending(&ct->timeout));

		WRITE_LOCK(&ip_conntrack_lock);
		if (++ip_conntrack_gen == UINT_MAX) {
			ip_conntrack_gen =
				   NF_REASON_MIN_RESERVED_FOR_CONNTRACK + 1;
		}

		/* Remove from both hash lists */
		LIST_DELETE(&ip_conntrack_hash
			    [hash_conntrack(&ct->tuplehash[IP_CT_DIR_ORIGINAL]
					    .tuple)],
			    &ct->tuplehash[IP_CT_DIR_ORIGINAL]);
		LIST_DELETE(&ip_conntrack_hash
			    [hash_conntrack(&ct->tuplehash[IP_CT_DIR_REPLY]
					    .tuple)],
			    &ct->tuplehash[IP_CT_DIR_REPLY]);
		/* If our expected is in the list, take it out. */
		if (list_inlist(&expect_list, &ct->expected))
			LIST_DELETE(&expect_list, &ct->expected);
		WRITE_UNLOCK(&ip_conntrack_lock);

		/* Need this outside the lock to avoid deadlock. */
		if (ip_conntrack_destroyed) {
			ip_conntrack_destroyed(ct);
		}

		kfree(ct);
	}
}

static void ip_conntrack_delete_timeout(unsigned long ul_conntrack)
{
	struct ip_conntrack *ct = (void *)ul_conntrack;

	IP_CONNTRACK_ASSERT(!ct->dead);

	ct->dead = 1;
	ip_conntrack_put(ct);
}

/* Returns TRUE if it dealt with ICMP (then needs ip_conntrack_put(h)
   if info not set to INVALID) */
int icmp_error_track_get(const struct iphdr *iph, size_t len,
			 enum ip_conntrack_info *info,
			 struct ip_conntrack_tuple_hash **h)
{
	struct icmphdr *hdr = (struct icmphdr *)((u_int32_t *)iph + iph->ihl);
	struct ip_conntrack_tuple innertuple, origtuple;
	struct iphdr *inner = (struct iphdr *)(hdr + 1);
	size_t datalen = len - iph->ihl*4 - sizeof(*hdr);
	struct ip_conntrack_protocol *innerproto;

	if (len < iph->ihl * 4 + sizeof(struct icmphdr)) {
		DEBUGP("icmp_error_track: too short\n");
		goto inval;
	}

	if (hdr->type != ICMP_DEST_UNREACH
	    && hdr->type != ICMP_SOURCE_QUENCH
	    && hdr->type != ICMP_TIME_EXCEEDED
	    && hdr->type != ICMP_PARAMETERPROB
	    && hdr->type != ICMP_REDIRECT)
		return 0;

	/* Ignore it if the checksum's bogus. */
	if (ip_compute_csum((unsigned char *)hdr, sizeof(*hdr) + datalen)) {
		DEBUGP("icmp_error_track: bad csum\n");
		goto inval;
	}

	innerproto = find_proto(inner->protocol);
	/* Are they talking about one of our connections? */
	if (inner->ihl * 4 + 8 > datalen
	    || !get_tuple(inner, datalen, &origtuple, innerproto)) {
		DEBUGP("icmp_error: ! get_tuple p=%u (%u*4+%u dlen=%u)\n",
		       inner->protocol, inner->ihl, 8,
		       datalen);
		goto inval;
	}

	/* Ordinarily, we'd expect the inverted tupleproto, but it's
	   been preserved inside the ICMP. */
	if (!invert_tuple(&innertuple, &origtuple, innerproto)) {
		DEBUGP("icmp_error_track: Can't invert tuple\n");
		goto inval;
	}
	*h = find_conntrack_get(&innertuple);
	if (!*h) {
		DEBUGP("icmp_error_track: no match\n");
		goto inval;
	}

	*info = IP_CT_RELATED;
	return 1;

 inval:
	*info = IP_CT_INVALID;
	return 1;
}

static int helper_cmp(const struct ip_conntrack_helper *i,
		      u_int16_t protonum,
		      const union ip_conntrack_manip_proto *repl)
{
	return i->protocol == protonum
		&& i->protocol_reply.all == repl->all;
}

/* Compare all but src per-proto part. */
static int expect_cmp(const struct ip_conntrack_expect *i,
		      const struct ip_conntrack_tuple *tuple)
{
	return (tuple->src.ip == i->tuple.src.ip
		&& tuple->dst.ip == i->tuple.dst.ip
		&& tuple->dst.u.all == i->tuple.dst.u.all
		&& tuple->dst.protonum == i->tuple.dst.protonum);
}

/* We can simply read off ip_conntrack_gen: if it's incremented, we
   know it's not because this conntrack was deleted (we have a
   reference). */
static void
update_skb(struct sk_buff *skb,
	   enum ip_conntrack_info info,
	   struct ip_conntrack_tuple_hash *h)
{
	IP_CONNTRACK_ASSERT(skb->nfreason == 0
			    || (skb->nfreason
				>= NF_REASON_MIN_RESERVED_FOR_CONNTRACK));
	if (info == IP_CT_INVALID) {
		skb->nfmark = 0;
		skb->nfreason = NF_REASON_MIN_RESERVED_FOR_CONNTRACK;
	} else {
		IP_CONNTRACK_ASSERT(h);
		IP_CONNTRACK_ASSERT(((unsigned long)h & 3) == 0);
		skb->nfmark = (unsigned long)h | info;
		skb->nfreason = ip_conntrack_gen;
	}
}

/* Allocate a new conntrack; we set everything up, then grab write
   lock and see if we lost a race.  If we lost it we simply cleanup
   and mark it INVALID (it should be rare, but FIXME). --RR */
static void
init_conntrack(const struct ip_conntrack_tuple *tuple,
	       struct ip_conntrack_protocol *protocol,
	       struct sk_buff *skb)
{
	struct ip_conntrack *conntrack = NULL;
	struct ip_conntrack_tuple repl_tuple;
	size_t hash, repl_hash;
	struct ip_conntrack_expect *expected = NULL;
	enum ip_conntrack_info ctinfo;
	struct ip_conntrack_helper *helper;

	if (!invert_tuple(&repl_tuple, tuple, protocol)) {	
		DEBUGP("Can't invert tuple.\n");
		ctinfo = IP_CT_INVALID;
		goto out;
	}

	conntrack = STRUCT_RESERVE_ALLOCATE(ip_conntrack, GFP_ATOMIC);
	if (!conntrack) {
		DEBUGP("Can't allocate conntrack.\n");
		ctinfo = IP_CT_INVALID;
		goto out;
	}
	hash = hash_conntrack(tuple);
	repl_hash = hash_conntrack(&repl_tuple);

	READ_LOCK(&ip_conntrack_lock);
	helper = LIST_FIND(&helpers, helper_cmp, struct ip_conntrack_helper *,
			   tuple->dst.protonum, &repl_tuple.src.u);
	READ_UNLOCK(&ip_conntrack_lock);

	*conntrack = ((struct ip_conntrack)
		      {{ {{ NULL, NULL }, *tuple, conntrack },
			 {{ NULL, NULL }, repl_tuple, conntrack },
		      },
		       1 << IPS_NORMAL, 0,
		       ATOMIC_INIT(1), /* home reference. */
		       { NULL, NULL, jiffies + protocol->expiry,
			 (unsigned long)conntrack, NULL },
		       /* Expected */
		       {{ NULL, NULL }, { { 0, { 0 }, 0 }, { 0, { 0 }, 0 } } },
		       /* helper */
		       helper });

	init_timer(&conntrack->timeout);
	conntrack->timeout.function = ip_conntrack_delete_timeout;
	add_timer(&conntrack->timeout);

	if (!protocol->new(conntrack, skb->nh.iph, skb->len))
		goto free_and_out;

	/* Sew in at head of hash list. */
	WRITE_LOCK(&ip_conntrack_lock);
	/* Check noone else beat us in the race... */
	if (LIST_FIND(&ip_conntrack_hash[hash], conntrack_cmp,
		      struct ip_conntrack_tuple_hash *, tuple)) {
		WRITE_UNLOCK(&ip_conntrack_lock);
		printk("ip_conntrack: Wow someone raced us!\n");
		goto free_and_out;
	}
	/* Need finding and deleting of expected ONLY if we win race */
	expected = LIST_FIND(&expect_list, expect_cmp,
			     struct ip_conntrack_expect *, tuple);
	if (expected) {
		/* Welcome, Mr. Bond.  We've been expecting you... */
		conntrack->status = (1 << IPS_EXPECTED);
		ctinfo = IP_CT_RELATED;
		LIST_DELETE(&expect_list, expected);
	} else {
		ctinfo = IP_CT_NEW;
	}
	list_prepend(&ip_conntrack_hash[hash],
		     &conntrack->tuplehash[IP_CT_DIR_ORIGINAL]);
	list_prepend(&ip_conntrack_hash[repl_hash],
		     &conntrack->tuplehash[IP_CT_DIR_REPLY]);
	WRITE_UNLOCK(&ip_conntrack_lock);

 out:
	update_skb(skb, ctinfo, &conntrack->tuplehash[IP_CT_DIR_ORIGINAL]);
	return;

 free_and_out:
	del_timer(&conntrack->timeout);
	kfree(conntrack);
	ctinfo = IP_CT_INVALID;
	goto out;
}

int
ip_conntrack_easy_get(struct sk_buff *skb, struct ip_conntrack_tuple_hash **h,
		      enum ip_conntrack_info *ctinfo)
{
	if (skb->nfreason
	    && skb->nfreason <= NF_REASON_MIN_RESERVED_FOR_CONNTRACK) {
		/* Not a generation count: being used by someone else,
		   or invalid. */
		DEBUGP("Skb %p INVALID reason = %u.\n", skb, skb->nfreason);
		*ctinfo = IP_CT_INVALID;
		return 1;
	}

	/* Need read lock to avoid race with gen count */
	READ_LOCK(&ip_conntrack_lock);
	if (skb->nfreason == ip_conntrack_gen) {
		enum ip_conntrack_info info = skb->nfmark & 3;

		*h = (struct ip_conntrack_tuple_hash *)(skb->nfmark & ~3UL);
		atomic_inc(&(*h)->ctrack->use);
		READ_UNLOCK(&ip_conntrack_lock);

		IP_CONNTRACK_ASSERT(info != IP_CT_NEW || DIRECTION(*h) == 0);
		*ctinfo = info;
		return 1;
	}
	READ_UNLOCK(&ip_conntrack_lock);
	return 0;
}

enum ip_conntrack_info
ip_conntrack_hard_get(struct sk_buff *skb,
		      const struct ip_conntrack_tuple *tuple,
		      struct ip_conntrack_tuple_hash **h)
{
	*h = find_conntrack_get(tuple);
	if (!*h)
		return IP_CT_INVALID;

	return skb->nfmark & 3;
}
/* Return conntrack_info and tuple hash given skb: we use the lower 2
   bits of the skb->nfmark to hold the type, and the rest to hold a
   pointer to a ip_conntrack_tuple_hash.  Gross, huh? */
enum ip_conntrack_info
ip_conntrack_get(struct sk_buff *skb, struct ip_conntrack_tuple_hash **h)
{
	enum ip_conntrack_info info;
	struct iphdr *iph = skb->nh.iph;
	struct ip_conntrack_tuple tuple;

	if (ip_conntrack_easy_get(skb, h, &info))
		return info;

	/* Generations don't match.  Look for it the long way. */
	if (iph->protocol == IPPROTO_ICMP
	    && icmp_error_track_get(iph, skb->len, &info, h)) {
		update_skb(skb, info, *h);
		return info;
	}

	if (!get_tuple(iph, skb->len, &tuple, find_proto(iph->protocol)))
		info = IP_CT_INVALID;
	else
		info = ip_conntrack_hard_get(skb, &tuple, h);

	update_skb(skb, info, *h);
	return info;
}

/* Netfilter hook itself. */
unsigned int ip_conntrack_in(unsigned int hooknum,
			     struct sk_buff **pskb,
			     const struct net_device *in,
			     const struct net_device *out)
{
	struct ip_conntrack_tuple_hash *h;
	struct iphdr *iph = (*pskb)->nh.iph;
	enum ip_conntrack_info info;
	struct ip_conntrack_protocol *proto;
	struct ip_conntrack_tuple tuple;
	unsigned int ret;

	/* FIXME: Do this right please. --RR */
	(*pskb)->nfcache |= NFC_UNKNOWN;

	/* Not for us, or previously seen (loopback)?  Ignore. */
	if ((*pskb)->nfreason)
		return NF_ACCEPT;

	/* Gather fragments. */	
	if (iph->frag_off & htons(IP_MF|IP_OFFSET)) {
#ifdef CONFIG_NETFILTER_DEBUG
		unsigned int olddebug = (*pskb)->nf_debug;
#endif
		*pskb = ip_defrag(*pskb);
		if (!*pskb)
			return NF_STOLEN;
		iph = (*pskb)->nh.iph;
		ip_send_check(iph);
		(*pskb)->nfcache |= NFC_ALTERED;
#ifdef CONFIG_NETFILTER_DEBUG
		/* Packet path as if nothing had happened. */
		(*pskb)->nf_debug = olddebug;
#endif
	}

	if (iph->protocol == IPPROTO_ICMP
	    && icmp_error_track_get(iph, (*pskb)->len, &info, &h)) {
		update_skb(*pskb, info, h);
		if (info != IP_CT_INVALID)
			ip_conntrack_put(h->ctrack);
		return NF_ACCEPT;
	}

	proto = find_proto(iph->protocol);
	if (!get_tuple(iph, (*pskb)->len, &tuple, proto)) {
		info = IP_CT_INVALID;
		return NF_ACCEPT;
	}

	h = find_conntrack_get(&tuple);
	if (!h) {
		init_conntrack(&tuple, proto, *pskb);
		return NF_ACCEPT;
	}

	/* It exists; we have (non-exclusive) reference. */
	if (DIRECTION(h)) {
		info = IP_CT_ESTABLISHED;
		set_bit(IPS_SEEN_REPLY, &h->ctrack->status);
	} else {
		/* Once we've had two way comms, all ESTABLISHED. */
		if (h->ctrack->status & (1 << IPS_SEEN_REPLY))
			info = IP_CT_ESTABLISHED;
		else if (h->ctrack->status & (1 << IPS_EXPECTED))
			info = IP_CT_RELATED;
		else {
			IP_CONNTRACK_ASSERT(h->ctrack->status
					    & (1 << IPS_NORMAL));
			info = IP_CT_NEW;
		}
	}

	ret = proto->established(h->ctrack, iph, (*pskb)->len, DIRECTION(h),
				 &info);
	if (ret != NF_DROP && info != IP_CT_INVALID) {
		struct ip_conntrack_helper *help = h->ctrack->helper;

		if (help)
			ret = help->help(iph, (*pskb)->len, h, &info);
	}
	update_skb(*pskb, info, h);
	ip_conntrack_put(h->ctrack);

	return ret;
}

static int
conntrack_tuple_cmp(const struct ip_conntrack_tuple_hash *i,
		    const struct ip_conntrack_tuple *tuple,
		    const struct ip_conntrack *ignored_conntrack)
{
	return !i->ctrack->dead
		&& i->ctrack != ignored_conntrack
		&& memcmp(tuple, &i->tuple, sizeof(*tuple)) == 0;
}

/* Find a connection corresponding to a tuple. */
struct ip_conntrack_tuple_hash *
ip_conntrack_find_get(const struct ip_conntrack_tuple *tuple,
		      const struct ip_conntrack *ignored_conntrack)
{
	struct ip_conntrack_tuple_hash *h;

	READ_LOCK(&ip_conntrack_lock);
	h = LIST_FIND(&ip_conntrack_hash[hash_conntrack(tuple)],
		      conntrack_tuple_cmp,
		      struct ip_conntrack_tuple_hash *,
		      tuple, ignored_conntrack);
	if (h)
		atomic_inc(&h->ctrack->use);
	READ_UNLOCK(&ip_conntrack_lock);

	return h;
}

int invert_tuplepr(struct ip_conntrack_tuple *inverse,
		   const struct ip_conntrack_tuple *orig)
{
	return invert_tuple(inverse, orig, find_proto(orig->dst.protonum));
}

/* Alter reply tuple (maybe alter helper). */
/* We're starting NAT: a race here in the reply field not really
   important, since noone else should have a reference to this
   conntrack */
void ip_conntrack_alter_reply(struct ip_conntrack *conntrack,
			      const struct ip_conntrack_tuple *newreply)
{
	unsigned int newindex = hash_conntrack(newreply);

	WRITE_LOCK(&ip_conntrack_lock);
	LIST_DELETE(&ip_conntrack_hash
		    [hash_conntrack(&conntrack->tuplehash[IP_CT_DIR_REPLY]
				    .tuple)],
		    &conntrack->tuplehash[IP_CT_DIR_REPLY]);
	conntrack->tuplehash[IP_CT_DIR_REPLY].tuple = *newreply;
	list_prepend(&ip_conntrack_hash[newindex],
		     &conntrack->tuplehash[IP_CT_DIR_REPLY]);
	conntrack->helper = LIST_FIND(&helpers, helper_cmp,
				      struct ip_conntrack_helper *,
				      newreply->dst.protonum,
				      &newreply->src.u);
	WRITE_UNLOCK(&ip_conntrack_lock);
}

int ip_conntrack_helper_register(struct ip_conntrack_helper *me)
{
	int ret;

	WRITE_LOCK(&ip_conntrack_lock);
	if (LIST_FIND(&helpers, helper_cmp,
		      struct ip_conntrack_helper *,
		      me->protocol, &me->protocol_reply)) {
		ret = -EBUSY;
	} else {
		ret = 0;
		list_prepend(&helpers, me);
		MOD_INC_USE_COUNT;
	}
	WRITE_UNLOCK(&ip_conntrack_lock);

	return ret;
}

static int unhelp(struct ip_conntrack_tuple_hash *i,
		  const struct ip_conntrack_helper *me)
{
	if (i->ctrack->helper == me) {
		i->ctrack->helper = NULL;
		/* Get rid of any expected. */
		if (list_inlist(&expect_list, &i->ctrack->expected))
			LIST_DELETE(&expect_list, &i->ctrack->expected);
	}
	return 0;
}

void ip_conntrack_helper_unregister(struct ip_conntrack_helper *me)
{
	unsigned int i;

	/* Need write lock here, to update dead. */
	WRITE_LOCK(&ip_conntrack_lock);
	LIST_DELETE(&helpers, me);

	/* Get rid of expecteds, set helpers to NULL. */
	for (i = 0; i < IP_CONNTRACK_HTABLE_SIZE; i++)
		LIST_FIND_W(&ip_conntrack_hash[i], unhelp,
			    struct ip_conntrack_tuple_hash *, me);
	WRITE_UNLOCK(&ip_conntrack_lock);

	/* Someone could be still looking at the helper in a bh.  Wait
           for them. */
	synchronize_bh();

	MOD_DEC_USE_COUNT;
}

/* Mishearing the voices in his head, our hero wonders how he's
   supposed to kill the mall. */
void ip_conntrack_cleanup(void)
{
	unsigned int i;
	unsigned long flags;

	/* Need to stop timers from deleting them here. */
	save_flags(flags);
	cli();

	/* Fairly inefficient, but it doesn't matter here. */
	/* If someone is holding one of these, we'll spin for a bit. */
	for (i = 0; i < IP_CONNTRACK_HTABLE_SIZE; i++) {
		READ_LOCK(&ip_conntrack_lock);
		while (!list_empty(&ip_conntrack_hash[i])) {
			struct ip_conntrack_tuple_hash *h
				= (void *)ip_conntrack_hash[i].next;

			if (!h->ctrack->dead) {
				atomic_inc(&h->ctrack->use);
				READ_UNLOCK(&ip_conntrack_lock);
				if (!del_timer(&h->ctrack->timeout))
					DEBUGP("Live conntrack %p no timer!\n",
					       h->ctrack);

				IP_CONNTRACK_ASSERT(!h->ctrack->dead);
				h->ctrack->dead = 1;
				ip_conntrack_put(h->ctrack);
				ip_conntrack_put(h->ctrack);
				READ_LOCK(&ip_conntrack_lock);
			}
		}
		READ_UNLOCK(&ip_conntrack_lock);
	}
	restore_flags(flags);
}

int ip_conntrack_init(void)
{
	unsigned int i;

	printk("ip_conntrack.o v%s\n", NETFILTER_VERSION);

	/* Don't NEED lock here, but good form anyway. */
	WRITE_LOCK(&ip_conntrack_lock);
	/* Sew in builtin protocols. */
	list_append(&protocol_list, &ip_conntrack_protocol_tcp);
	list_append(&protocol_list, &ip_conntrack_protocol_udp);
	list_append(&protocol_list, &ip_conntrack_protocol_icmp);

	for (i = 0; i < IP_CONNTRACK_HTABLE_SIZE; i++)
		INIT_LIST_HEAD(&ip_conntrack_hash[i]);
	WRITE_UNLOCK(&ip_conntrack_lock);

	return ip_conntrack_protocol_tcp_init();
}
