/* This file contains all the functions required for the standalone
   ip_conntrack module.

   These are not required by the compatibility layer.
*/

/* (c) 1999 Paul `Rusty' Russell.  Licenced under the GNU General
   Public Licence. */

#define EXPORT_SYMTAB
#include <linux/config.h>
#include <linux/types.h>
#include <linux/ip.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <net/checksum.h>

#define ASSERT_READ_LOCK(x) MUST_BE_READ_LOCKED(&ip_conntrack_lock)
#define ASSERT_WRITE_LOCK(x) MUST_BE_WRITE_LOCKED(&ip_conntrack_lock)

#include "conntrack/struct_reserve.h"
#include "conntrack/ip_conntrack.h"
#include "conntrack/ip_conntrack_protocol.h"
#include "conntrack/ip_conntrack_core.h"
#include "conntrack/ip_conntrack_helper.h"
#include "common/listhelp.h"

struct module *ip_conntrack_module = THIS_MODULE;

static unsigned int
print_tuple(char *buffer, const struct ip_conntrack_tuple *tuple,
	    struct ip_conntrack_protocol *proto)
{
	int len;

	len = sprintf(buffer, "src=%u.%u.%u.%u dst=%u.%u.%u.%u ",
		      IP_PARTS(tuple->src.ip), IP_PARTS(tuple->dst.ip));

	len += proto->print_tuple(buffer + len, tuple);

	return len;
}

/* FIXME: Don't print source proto part. --RR */
static unsigned int
print_expect(char *buffer, const struct ip_conntrack_expect *expect)
{
	unsigned int len;

	len = sprintf(buffer, "EXPECTING: proto=%u ",
		      expect->tuple.dst.protonum);
	len += print_tuple(buffer + len, &expect->tuple,
			   __find_proto(expect->tuple.dst.protonum));
	len += sprintf(buffer + len, "\n");
	return len;
}

static unsigned int
print_conntrack(char *buffer, const struct ip_conntrack *conntrack)
{
	unsigned int len;
	struct ip_conntrack_protocol *proto
		= __find_proto(conntrack->tuplehash[IP_CT_DIR_ORIGINAL]
			       .tuple.dst.protonum);

	len = sprintf(buffer, "%-8s %u %lu ",
		      proto->name,
		      conntrack->tuplehash[IP_CT_DIR_ORIGINAL]
		      .tuple.dst.protonum,
		      timer_pending(&conntrack->timeout)
		      ? (conntrack->timeout.expires - jiffies)/HZ : 0);

	len += proto->print_conntrack(buffer + len, conntrack);
	len += print_tuple(buffer + len,
			   &conntrack->tuplehash[IP_CT_DIR_ORIGINAL].tuple,
			   proto);
	if (!(conntrack->status & (1 << IPS_SEEN_REPLY)))
		len += sprintf(buffer + len, "[UNREPLIED] ");
	len += print_tuple(buffer + len,
			   &conntrack->tuplehash[IP_CT_DIR_REPLY].tuple,
			   proto);
	len += sprintf(buffer + len, "\n");

	return len;
}

/* Returns true when finished. */
static int
conntrack_iterate(const struct ip_conntrack_tuple_hash *hash,
		  char *buffer, off_t offset, off_t *upto,
		  unsigned int *len, unsigned int maxlen)
{
	unsigned int newlen;
	IP_CONNTRACK_ASSERT(hash->ctrack);

	/* Don't print if dead. */
	if (hash->ctrack->dead)
		return 0;

	/* Only count originals */
	if (DIRECTION(hash))
		return 0;

	if ((*upto)++ < offset)
		return 0;

	newlen = print_conntrack(buffer + *len, hash->ctrack);
	if (*len + newlen > maxlen)
		return 1;
	else *len += newlen;

	return 0;
}

static int
list_conntracks(char *buffer, char **start, off_t offset, int length,
		int dummy)
{
	unsigned int i;
	unsigned int len = 0;
	off_t upto = 0;
	struct list_head *e;

	READ_LOCK(&ip_conntrack_lock);
	/* Traverse hash; print originals then reply. */
	for (i = 0; i < IP_CONNTRACK_HTABLE_SIZE; i++) {
		if (LIST_FIND(&ip_conntrack_hash[i], conntrack_iterate,
			      struct ip_conntrack_tuple_hash *,
			      buffer, offset, &upto, &len, length))
			goto finished;
	}

	/* Now iterate through expecteds. */
	for (e = expect_list.next; e != &expect_list; e = e->next) {
		unsigned int last_len;
		struct ip_conntrack_expect *expect
			= (struct ip_conntrack_expect *)e;
		if (upto++ < offset) continue;

		last_len = len;
		len += print_expect(buffer + len, expect);
		if (len > length) {
			len = last_len;
			goto finished;
		}
	}

 finished:
	READ_UNLOCK(&ip_conntrack_lock);

	/* `start' hack - see fs/proc/generic.c line ~165 */
	*start = (char *)((unsigned int)upto - offset);
	return len;
}

#ifdef CONFIG_NETFILTER_DEBUG
static int fix_flags(struct sk_buff *skb)
{
	skb->nf_debug = (1 << NF_IP_LOCAL_OUT);
	return skb->dst->output(skb);
}
#endif

static unsigned int
ip_conntrack_local(unsigned int hooknum,
		   struct sk_buff **pskb,
		   const struct net_device *in,
		   const struct net_device *out)
{	
	unsigned int ret;
	int need_refrag = ((*pskb)->nh.iph->frag_off & htons(IP_MF|IP_OFFSET));
#ifdef CONFIG_NETFILTER_DEBUG
	struct sock *sk = (*pskb)->sk;
#endif

	ret = ip_conntrack_in(hooknum, pskb, in, out);
	if (ret != NF_STOLEN && need_refrag) {
		/* It defragmented.  Fragment it again, like they expect. */
#ifdef CONFIG_NETFILTER_DEBUG
		(*pskb)->sk = sk;
		ip_fragment(*pskb, fix_flags);
#else
		ip_fragment(*pskb, (*pskb)->dst->output);
#endif
		ret = NF_STOLEN;
	}
	return ret;
}

/* Connection tracking may drop packets, but never alters them, so
   make it the first hook. */
static struct nf_hook_ops ip_conntrack_in_ops
= { { NULL, NULL }, ip_conntrack_in, NULL, PF_INET, NF_IP_PRE_ROUTING, -200 };
static struct nf_hook_ops ip_conntrack_local_out_ops
= { { NULL, NULL }, ip_conntrack_local, NULL, PF_INET, NF_IP_LOCAL_OUT, -200 };

#ifdef CONFIG_PROC_FS
struct proc_dir_entry ip_conntrack_procinfo = {
	0, 12, "ip_conntrack", S_IFREG | S_IRUGO,
	1, 0, 0, 0, &proc_net_inode_operations, &list_conntracks
};
#endif

static int init_or_cleanup(int init)
{
	int ret = 0;

	if (!init) goto cleanup;

	ret = ip_conntrack_init();
	if (ret < 0)
		goto cleanup_nothing;

#ifdef CONFIG_PROC_FS
	proc_net_register(&ip_conntrack_procinfo);
#endif
	ret = nf_register_hook(&ip_conntrack_in_ops);
	if (ret < 0) {
		printk("ip_conntrack_init: can't register in hook.\n");
		goto cleanup_init;
	}
	ret = nf_register_hook(&ip_conntrack_local_out_ops);
	if (ret < 0) {
		printk("ip_conntrack_init: can't register local out hook.\n");
		goto cleanup_inops;
	}

	return ret;

 cleanup:
	nf_unregister_hook(&ip_conntrack_local_out_ops);
	
 cleanup_inops:
	nf_unregister_hook(&ip_conntrack_in_ops);
 cleanup_init:
#ifdef CONFIG_PROC_FS
	proc_net_unregister(ip_conntrack_procinfo.low_ino);
#endif
	ip_conntrack_cleanup();
 cleanup_nothing:
	return ret;
}

/* FIXME: Allow NULL functions and sub in pointers to generic for
   them. --RR */
int ip_conntrack_protocol_register(struct ip_conntrack_protocol *proto)
{
	int ret = 0;
	struct list_head *i;

	WRITE_LOCK(&ip_conntrack_lock);
	for (i = protocol_list.next; i != &protocol_list; i = i->next) {
		if (((struct ip_conntrack_protocol *)i)->proto
		    == proto->proto) {
			ret = -EBUSY;
			goto out;
		}
	}

	list_prepend(&protocol_list, proto);
	MOD_INC_USE_COUNT;

 out:
	WRITE_UNLOCK(&ip_conntrack_lock);
	return ret;
}

/* FIXME: This is completely wrong. --RR */
#if 0
static inline int
proto_cmp(const struct ip_conntrack_tuple_hash *h, u_int16_t protonum)
{
	return !h->ctrack->dead && (h->tuple.dst.protonum == protonum);
}

static struct ip_conntrack_tuple_hash *
get_hash_by_proto(u_int16_t protonum)
{
	struct ip_conntrack_tuple_hash *h = NULL;
	size_t i;

	READ_LOCK(&ip_conntrack_lock);
	for (i = 0; !h && i < IP_CONNTRACK_HTABLE_SIZE; i++) {
		h = LIST_FIND(&ip_conntrack_hash[i], proto_cmp,
			      struct ip_conntrack_tuple_hash *, protonum);
	}
	if (h)
		atomic_inc(&h->ctrack->use);
	READ_UNLOCK(&ip_conntrack_lock);

	return h;
}

void ip_conntrack_protocol_unregister(struct ip_conntrack_protocol *proto)
{
	struct ip_conntrack_tuple_hash *h;

	WRITE_LOCK(&ip_conntrack_lock);
	LIST_DELETE(&protocol_list, proto);
	WRITE_UNLOCK(&ip_conntrack_lock);

	/* Noone else will use protocol after this. */
	synchronize_bh();

	/* This is hugely inefficient (n^2), but it doesn't matter,
	   since removing a module isn't speed critical, and this
	   makes locking simpler. */
	while ((h = get_hash_by_proto(proto->proto)) != NULL) {
		if (del_timer(&h->ctrack->timeout)) {
			IP_CONNTRACK_ASSERT(!h->ctrack->dead);
			h->ctrack->dead = 1;
			ip_conntrack_put(h->ctrack);
		}
		ip_conntrack_put(h->ctrack);
	}
	MOD_DEC_USE_COUNT;
}
#endif

int init_module(void)
{
	return init_or_cleanup(1);
}

void cleanup_module(void)
{
	init_or_cleanup(0);
}

EXPORT_SYMBOL(ip_conntrack_protocol_register);
#if 0
EXPORT_SYMBOL(ip_conntrack_protocol_unregister);
#endif
EXPORT_SYMBOL(expect_list);
EXPORT_SYMBOL(invert_tuplepr);
EXPORT_SYMBOL(ip_conntrack_alter_reply);
EXPORT_SYMBOL(ip_conntrack_destroyed);
EXPORT_SYMBOL(ip_conntrack_get);
EXPORT_SYMBOL(ip_conntrack_module);
EXPORT_STRUCT_RESERVABLE(ip_conntrack);
EXPORT_SYMBOL(ip_conntrack_helper_register);
EXPORT_SYMBOL(ip_conntrack_helper_unregister);
EXPORT_SYMBOL(ip_conntrack_find_get);
EXPORT_SYMBOL(ip_conntrack_put);
EXPORT_SYMBOL(ip_conntrack_hash);
EXPORT_SYMBOL(ip_conntrack_lock);
EXPORT_SYMBOL(ip_conntrack_hard_get);
EXPORT_SYMBOL(ip_conntrack_easy_get);
EXPORT_SYMBOL(find_proto);
EXPORT_SYMBOL(get_tuple);
