#include <linux/types.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/netfilter.h>
#include <linux/icmp.h>
#include "conntrack/ip_conntrack.h"

#ifdef FAST_TIMEOUTS
#define ICMP_TIMEOUT (10*HZ)
#else
#define ICMP_TIMEOUT (30*HZ)
#endif

static int icmp_pkt_to_tuple(const void *datah, size_t datalen,
			     struct ip_conntrack_tuple *tuple)
{
	const struct icmphdr *hdr = datah;

	tuple->dst.u.icmp.type = hdr->type;
	tuple->src.u.icmp.id = hdr->un.echo.id;
	tuple->dst.u.icmp.code = hdr->code;

	return 1;
}

static int icmp_invert_tuple(struct ip_conntrack_tuple *tuple,
			     const struct ip_conntrack_tuple *orig)
{
	/* Add 1; spaces filled with 0. */
	static u_int8_t invmap[]
		= { [ICMP_ECHO] = ICMP_ECHOREPLY + 1,
		    [ICMP_ECHOREPLY] = ICMP_ECHO + 1,
		    [ICMP_TIMESTAMP] = ICMP_TIMESTAMPREPLY + 1,
		    [ICMP_TIMESTAMPREPLY] = ICMP_TIMESTAMP + 1,
		    [ICMP_INFO_REQUEST] = ICMP_INFO_REPLY + 1,
		    [ICMP_INFO_REPLY] = ICMP_INFO_REQUEST + 1,
		    [ICMP_ADDRESS] = ICMP_ADDRESSREPLY + 1,
		    [ICMP_ADDRESSREPLY] = ICMP_ADDRESS + 1};

	if (orig->dst.u.icmp.type >= sizeof(invmap)
	    || !invmap[orig->dst.u.icmp.type])
		return 0;

	tuple->src.u.icmp.id = orig->src.u.icmp.id;
	tuple->dst.u.icmp.type = invmap[orig->dst.u.icmp.type] - 1;
	tuple->dst.u.icmp.code = orig->dst.u.icmp.code;
	return 1;
}

/* Print out the per-protocol part of the tuple. */
static unsigned int icmp_print_tuple(char *buffer,
				     const struct ip_conntrack_tuple *tuple)
{
	return sprintf(buffer, "type=%u code=%u id=%u ",
		       tuple->dst.u.icmp.type,
		       tuple->dst.u.icmp.code,
		       ntohs(tuple->src.u.icmp.id));
}

/* Print out the private part of the conntrack. */
static unsigned int icmp_print_conntrack(char *buffer,
				     const struct ip_conntrack *conntrack)
{
	return 0;
}

/* Returns verdict for packet, and may modify conntracktype */
static unsigned int icmp_established(struct ip_conntrack *conntrack,
				     struct iphdr *iph, size_t len,
				     enum ip_conntrack_dir direction,
				     enum ip_conntrack_info *conntrackinfo)
{
	WRITE_LOCK(&ip_conntrack_lock);
	/* FIXME: If we delete this conntrack, then NAT can't access it.
	   So we simply delete it "soon". --RR */
	if (del_timer(&conntrack->timeout)) {
		conntrack->timeout.expires = jiffies + 2;
		add_timer(&conntrack->timeout);
	}
	WRITE_UNLOCK(&ip_conntrack_lock);

	return NF_ACCEPT;
}

/* Called when a new connection for this protocol found. */
static int icmp_new(struct ip_conntrack *conntrack,
		     struct iphdr *iph, size_t len)
{
	static u_int8_t valid_new[]
		= { [ICMP_ECHO] = 1,
		    [ICMP_TIMESTAMP] = 1,
		    [ICMP_INFO_REQUEST] = 1,
		    [ICMP_ADDRESS] = 1 };

	if (conntrack->tuplehash[0].tuple.dst.u.icmp.type >= sizeof(valid_new)
	    || !valid_new[conntrack->tuplehash[0].tuple.dst.u.icmp.type]) {
		/* Can't create a new ICMP `conn' with this. */
		DEBUGP("icmp: can't create new conn with type %u\n",
		       conntrack->tuplehash[0].tuple.dst.u.icmp.type);
		return 0;
	}
	return 1;
}

struct ip_conntrack_protocol ip_conntrack_protocol_icmp
= { { NULL, NULL }, IPPROTO_ICMP, "icmp",
    icmp_pkt_to_tuple, icmp_invert_tuple, icmp_print_tuple,
    icmp_print_conntrack, icmp_established, icmp_new, NULL,
    ICMP_TIMEOUT };
