/* Generic functions to manage a system where modules want to reserve
   private space in your structure. */

#ifndef _STRUCT_RESERVE_H
#define _STRUCT_RESERVE_H

#define _STRUCT_CAT(x,y) x ## y

struct reserve_struct
{
	long gen;
	size_t offset;
};

/* Hand these the type name, the space you require, and the alignment
   (use __alignof__()).  Fills in structure pointer, returns void. */
#define struct_reserve(type, space, align, rstrp)			     \
	__struct_reserve(&_STRUCT_CAT(__reserve_,type), space, align, rstrp)

/* Returns void. */
#define struct_unreserve(type, space, rstrp)				  \
	__struct_unreserve(&_STRUCT_CAT(__reserve_,type), space, rstrp)

#define EXTERN_DECLARE_STRUCT_RESERVABLE(type)		\
	extern struct reserve_struct _STRUCT_CAT(__reserve_,type)

#define DECLARE_STRUCT_RESERVABLE(type)				\
struct reserve_struct _STRUCT_CAT(__reserve_,type)		\
= { 0, __ALIGN_UP(sizeof(struct type), __alignof__(long)) + sizeof(long) }

#define EXPORT_STRUCT_RESERVABLE(type)			\
	EXPORT_SYMBOL(_STRUCT_CAT(__reserve_,type))

/* We use a master reserve struct to monitor current size and generation */
#define STRUCT_RESERVE_ALLOCATE(type, pri)				    \
({ size_t __s = _STRUCT_CAT(__reserve_,type).offset;			    \
 void *__x = kmalloc(__s, pri);						    \
 if (__x) {								    \
	 size_t __off = __ALIGN_UP(sizeof(struct type),	__alignof__(long)); \
	 *(long *)(__x + __off)	= _STRUCT_CAT(__reserve_,type).gen;	    \
	 memset(__x + __off + sizeof(long), 0, __s - sizeof(long) - __off); \
 }									    \
 __x;									    \
})

#define STRUCT_RESERVE_GEN(type, ptr)				\
(*(long *)((void *)(ptr) + (__ALIGN_UP(sizeof(struct type),	\
				       __alignof__(long)))))

#ifdef DEBUG_CONNTRACK
#define STRUCT_RESERVED_PTR(type, structp, rstrp)			    \
({ void *__p;								    \
 if (STRUCT_RESERVE_GEN(type, structp) - (rstrp)->gen >= 0)		    \
	 __p = (void *)(structp) + (rstrp)->offset;			    \
 else {									    \
	 printk("%s:%u gen %lu < %lu\n", __FILE__, __LINE__,		    \
		STRUCT_RESERVE_GEN(type, structp), (rstrp)->gen);	    \
	 __p = NULL;							    \
 }									    \
 __p;									    \
})
#else
#define STRUCT_RESERVED_PTR(type, structp, rstrp)			   \
(STRUCT_RESERVE_GEN(type, structp) - (rstrp)->gen >= 0			   \
 ? ((void *)(structp) + (rstrp)->offset) : NULL)
#endif

/*** Internals below this line. ***/

#define __ALIGN_UP(x, a) (((x) + (a) - 1) & ~((a)-1))

/* Must not be called reentrant, or at same time as __struct_unreserve. */
extern inline void
__struct_reserve(struct reserve_struct *master,
		 size_t size,
		 size_t align,
		 struct reserve_struct *rstrp)
{
#ifdef DEBUG_CONNTRACK
	printk("struct_reserve: size %u align %u\n", size, align);
#endif
	/* Simply tack it on the end. */
	rstrp->offset = __ALIGN_UP(master->offset, align);
#ifdef DEBUG_CONNTRACK
	printk("struct_reserve: total size from %u to %u (offset = %u)\n",
	       master->offset, rstrp->offset + size, rstrp->offset);
#endif
	master->offset = rstrp->offset + size;
	rstrp->gen = ++master->gen;
}

extern inline void
__struct_unreserve(struct reserve_struct *master,
		   size_t size,
		   struct reserve_struct *rstrp)
{
	/* Trim if at end.  (Doesn't trim alignment padding.) */
	if (size + rstrp->offset == master->offset)
		master->offset -= size;
}
#endif /*_STRUCT_RESERVE_H*/
