/*
 * 25-Jul-1998 Major changes to allow for ip chain table
 */

/*
 * 	Format of an IP firewall descriptor
 *
 * 	src, dst, src_mask, dst_mask are always stored in network byte order.
 * 	flags are stored in host byte order (of course).
 * 	Port numbers are stored in HOST byte order.
 */

#ifndef _IP_TABLES_H
#define _IP_TABLES_H

#ifdef __KERNEL__
#include <linux/if.h>
#include <linux/types.h>
#include <linux/in.h>
#include <linux/ip.h>
#include <linux/skbuff.h>
#endif
#include <linux/netfilter_ipv4.h>

#define IPT_FUNCTION_MAXNAMELEN 16

struct ipt_ip {
	/* Source and destination IP addr */
	struct in_addr src, dst;		
	/* Mask for src and dest IP addr */
	struct in_addr smsk, dmsk;
	char iniface[IFNAMSIZ], outiface[IFNAMSIZ];
	unsigned char iniface_mask[IFNAMSIZ], outiface_mask[IFNAMSIZ];

	/* Protocol, 0 = ANY */
	u_int16_t proto;
	/* TOS to match iff flags & IPT_F_TOS */
	u_int8_t tos;

	/* Flags word */
	u_int8_t flags;
	/* Inverse flags */
	u_int8_t invflags;
};

/* tcp specific stuff */
struct ipt_tcp
{
	u_int16_t spts[2];			/* Source port range. */
	u_int16_t dpts[2];			/* Destination port range. */
	u_int8_t option;			/* TCP Option iff non-zero*/
	u_int8_t flg_mask;			/* TCP flags mask byte */
	u_int8_t flg_cmp;			/* TCP flags compare byte */
	u_int8_t invflags;			/* Inverse flags */
};

/* udp specific stuff */
struct ipt_udp
{
	u_int16_t spts[2];			/* Source port range. */
	u_int16_t dpts[2];			/* Destination port range. */
	u_int8_t invflags;			/* Inverse flags */
};

/* icmp specific stuff */
struct ipt_icmp
{
	u_int8_t type;				/* type to match */
	u_int8_t code[2];			/* range of code */
	u_int8_t invflags;			/* Inverse flags */
};

struct ipt_counters
{
	u_int64_t pcnt, bcnt;			/* Packet and byte counters */
};

union ipt_matchinfo
{
	struct ipt_tcp tcp;
	struct ipt_udp udp;
	struct ipt_icmp icmp;
	unsigned char other[16];
};

/* This is defined here for convenience; it also sets the size
   available for other targets */
struct ipt_standard
{
	int verdict;		/* What to do with packet: position or
				 * -(RESULT) - 1. */
	u_int32_t padding[3];	/* 16 bytes gives more room for others. */
};

union ipt_targinfo
{
	struct ipt_standard standard;
	char error[sizeof(struct ipt_standard)];
	unsigned char other[sizeof(struct ipt_standard)];
};

/* This structure defines each of the firewall rules.  Consists of 3
   parts which are 1) general IP header stuff 2) match specific
   stuff 3) the target to perform if the rule matches */
struct ipt_entry
{
	struct ipt_ip ip;
	union ipt_matchinfo matchinfo;

	/* Mark with fields that we care about. */
	unsigned int nfcache;

	/* Name of match function. */
	char match_name[IPT_FUNCTION_MAXNAMELEN];
	/* Name of target to take. */
	char target_name[IPT_FUNCTION_MAXNAMELEN];
	/* Packet and byte counters. */
	struct ipt_counters counters;

	union ipt_targinfo targinfo;
};

/* Values for "flag" field in struct ipt_ip (general ip structure). */
#define IPT_F_FRAG		0x01	/* Set if rule is a fragment rule */
#define IPT_F_TOS		0x02	/* Match the TOS. */

#define IPT_F_MASK		0x03	/* All possible flag bits mask. */

/* Values for "inv" field in struct ipt_ip. */
#define IPT_INV_VIA_IN		0x01	/* Invert the sense of IN IFACE. */
#define IPT_INV_VIA_OUT		0x02	/* Invert the sense of OUT IFACE */
#define IPT_INV_TOS		0x04	/* Invert the sense of TOS. */
#define IPT_INV_SRCIP		0x08	/* Invert the sense of SRC IP. */
#define IPT_INV_DSTIP		0x10	/* Invert the sense of DST OP. */
#define IPT_INV_FRAG		0x20	/* Invert the sense of FRAG. */
#define IPT_INV_PROTO		0x40	/* Invert the sense of PROTO. */
#define IPT_INV_MASK		0x7F	/* All possible flag bits mask. */

/*
 *	New IP firewall options for [gs]etsockopt at the RAW IP level.
 *	Unlike BSD Linux inherits IP options so you don't have to use
 * a raw socket for this. Instead we check rights in the calls. */

#define IPT_BASE_CTL		64	/* base for firewall socket options */

#define IPT_SO_SET_REPLACE	(IPT_BASE_CTL)
#define IPT_SO_SET_ADD_COUNTERS	(IPT_BASE_CTL + 1)
#define IPT_SO_SET_MAX		IPT_SO_SET_ADD_COUNTERS

#define IPT_SO_GET_INFO		(IPT_BASE_CTL)
#define IPT_SO_GET_ENTRIES	(IPT_BASE_CTL + 1)
#define IPT_SO_GET_VERSION	(IPT_BASE_CTL + 2)
#define IPT_SO_GET_MAX		IPT_SO_GET_VERSION

/* RETURN verdict */
#define IPT_RETURN (NF_MAX_VERDICT+1)

#define IPT_HOOK_INPUT (NF_IP_LOCAL_IN - 1)
#define IPT_HOOK_FORWARD (NF_IP_FORWARD - 1)
#define IPT_HOOK_OUTPUT (NF_IP_LOCAL_OUT - 1)

/* The argument to IPT_SO_GET_INFO */
struct ipt_getinfo
{
	/* Hook entry points: INPUT, FORWARD and OUTPUT */
	unsigned int hook_entry[3];

	/* Underflow points: INPUT, FORWARD and OUTPUT */
	unsigned int underflow[3];

	/* Number of entries. */
	unsigned int num_entries;
};

/* The argument to IPT_SO_REPLACE. */
struct ipt_replace
{
	/* Number of entries to be inserted. */
	unsigned int number;

	/* Hook entry points. */
	unsigned int hook_entry[3];

	/* Underflow points: INPUT, FORWARD and OUTPUT */
	unsigned int underflow[3];

	/* Information about entries' counters to saved: */
	/* Number of entries (must be equal to current number of entries). */
	unsigned int num_counters;
	/* The counters. */
	struct ipt_counters *counters;

	/* The entries (hang off end) */
	struct ipt_entry entry_array[0];
};

/* The argument to IPT_SO_ADD_COUNTERS. */
struct ipt_counters_info
{
	unsigned int num_counters;

	/* The counters (actually `number' of these). */
	struct ipt_counters counters[0];
};

/* The argument to IPT_SO_GET_ENTRIES. */
struct ipt_get_entries
{
	unsigned int num_entries;

	/* The entries (actually `number' of these). */
	struct ipt_entry entries[0];
};

/* Standard return verdict, or do jump. */
#define IPT_STANDARD_TARGET ""
#define IPT_STANDARD_MATCH ""

/* Error verdict. */
#define IPT_ERROR_TARGET "ERROR"

/*
 *	Main firewall chains definitions and global var's definitions.
 */

#if defined(__KERNEL__) || defined (IP_TABLES_USER)

#include <linux/config.h>
#include <linux/init.h>
extern void ipt_init(void) __init;

struct ipt_match
{
	struct list_head list;

	const char name[IPT_FUNCTION_MAXNAMELEN];

	const char *version;

	/* Return true or false: return FALSE and set *hotdrop = 1 to
           force immediate packet drop. */
	int (*match)(struct sk_buff *skb,
		     const char *indev,
		     const char *outdev,
		     const union ipt_matchinfo *matchinfo,
		     int offset,
		     const void *hdr,
		     u_int16_t datalen,
		     int *hotdrop);
	/* Called when user tries to insert an entry of this type. */
	/* Should return true or false. */
	int (*checkentry)(const struct ipt_ip *ip,
			  const union ipt_matchinfo *matchinfo,
			  unsigned int chain_mask);

	/* Called when user tries to delete an entry of this type. */
	void (*deleteentry)(const union ipt_matchinfo *info);

	/* Set this to THIS_MODULE if you are a module, otherwise NULL */
	struct module *me;
};

/* Registration hooks for targets. */
struct ipt_target
{
	struct list_head list;

	const char name[IPT_FUNCTION_MAXNAMELEN];

	const char *version;

	/* Return -verdict - 1 or a position. */
	int (*target)(struct sk_buff *skb,
		      const char *indev,
		      const char *outdev,
		      const union ipt_targinfo *info,
		      unsigned int pos);
	/* Called when user tries to insert an entry of this type. */
	/* Should return true or false. */
	int (*checkentry)(const union ipt_targinfo *info,
			  unsigned int pos,
			  unsigned int finalpos,
			  unsigned int chain_mask);

	/* Called when user tries to delete an entry of this type. */
	void (*deleteentry)(const union ipt_targinfo *info);

	/* Set this to THIS_MODULE if you are a module, otherwise NULL */
	struct module *me;
};

int ipt_register_target(struct ipt_target *target);
void ipt_unregister_target(struct ipt_target *target);

int ipt_register_match(struct ipt_match *match);
void ipt_unregister_match(struct ipt_match *match);
#endif /*__KERNEL__*/
#endif /* _IP_TABLES_H */
