/* Kernel module to match UDP parameters. */
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/udp.h>

#include "packet-filter/extensions/ipt_udp.h"
#include "packet-filter/kernel/ip_tables.h"

EXPORT_NO_SYMBOLS;

#ifdef DEBUG_CONNTRACK
#define duprintf(format, args...) printk(format , ## args)
#else
#define duprintf(format, args...)
#endif

/* Returns 1 if the port is matched by the range, 0 otherwise */
static inline int
port_match(u_int16_t min, u_int16_t max, u_int16_t port, int invert)
{
	return (port >= min && port <= max) ^ invert;
}

static int
match(struct sk_buff *skb,
      const char *indev,
      const char *outdev,
      const union ipt_matchinfo *matchinfo,
      int offset,
      const void *hdr,
      u_int16_t datalen,
      int *hotdrop)
{
	const struct udphdr *udp = hdr;

	if (offset == 0 && datalen < sizeof(struct udphdr)) {
		/* We've been asked to examine this packet, and we
		   can't.  Hence, no choice but to drop. */
		duprintf("Dropping evil UDP tinygram.\n");
		*hotdrop = 1;
		return 0;
	}

	/* Must not be a fragment. */
	return !offset
		&& port_match(matchinfo->udp.spts[0], matchinfo->udp.spts[1],
			      ntohs(udp->source),
			      !!(matchinfo->udp.invflags & IPT_UDP_INV_SRCPT))
		&& port_match(matchinfo->udp.dpts[0], matchinfo->udp.dpts[1],
			      ntohs(udp->dest),
			      !!(matchinfo->udp.invflags & IPT_UDP_INV_DSTPT));
}

/* Called when user tries to insert an entry of this type. */
static int
checkentry(const struct ipt_ip *ip,
	   const union ipt_matchinfo *matchinfo,
	   unsigned int chain_mask)
{
	/* Must specify proto == UDP, and no unknown invflags */
	return ip->proto == IPPROTO_UDP
		&& !(matchinfo->udp.invflags & ~IPT_UDP_INV_MASK);
}

static struct ipt_match udp_match
= { { NULL, NULL }, "udp", NETFILTER_VERSION,
    &match, &checkentry, NULL, THIS_MODULE };

int __init init(void)
{
	return ipt_register_match(&udp_match);
}

void __exit cleanup(void)
{
	ipt_unregister_match(&udp_match);
}

module_init(init);
module_exit(cleanup);
