/* Shared library add-on to iptables to add MAC address support. */
#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#if defined(__GLIBC__) && __GLIBC__ == 2
#include <net/ethernet.h>
#else
#include <linux/if_ether.h>
#endif
#include "packet-filter/userspace/iptables.h"
#include "packet-filter/extensions/ipt_mac.h"

/* Function which prints out usage message. */
static void
help(void)
{
	printf(
"MAC v%s options:\n"
" --mac-source [!] XX:XX:XX:XX:XX:XX\n"
"				Match source MAC address\n"
"\n", NETFILTER_VERSION);
}

static struct option opts[] = {
	{ "mac-source", 1, 0, '1' },
	{0}
};

static void
init(union ipt_matchinfo *matchinfo, unsigned int *nfcache)
{
	/* Can't cache this. */
	*nfcache |= NFC_UNKNOWN;
}

static void
parse_mac(const char *mac, struct ipt_mac_info *info)
{
	unsigned int i = 0;

	if (strlen(mac) != ETH_ALEN*3-1)
		exit_error(PARAMETER_PROBLEM, "Bad mac address `%s'", mac);

	for (i = 0; i < ETH_ALEN; i++) {
		long number;
		char *end;

		number = strtol(mac + i*3, &end, 16);

		if (end == mac + i*3 + 2
		    && number >= 0
		    && number <= 255)
			info->srcaddr[i] = number;
		else
			exit_error(PARAMETER_PROBLEM,
				   "Bad mac address `%s'", mac);
	}
}

/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
      struct ipt_entry *entry)
{
	struct ipt_mac_info *mac
		= (struct ipt_mac_info *)&entry->matchinfo.other;
	switch (c) {
	case '1':
		if (check_inverse(optarg, &invert))
			optind++;
		parse_mac(argv[optind-1], mac);
		if (invert)
			mac->invert = 1;
		*flags = 1;
		break;

	default:
		return 0;
	}

	return 1;
}

static void print_mac(unsigned char macaddress[ETH_ALEN], int invert)
{
	unsigned int i;

	printf("%s%02X", invert ? "!" : "", macaddress[0]);
	for (i = 1; i < ETH_ALEN; i++)
		printf(":%02X", macaddress[i]);
}

/* Final check; must have specified --mac. */
static void final_check(unsigned int flags)
{
	if (!flags)
		exit_error(PARAMETER_PROBLEM,
			   "You must specify `--mac-source'");
}

/* Prints out the union ipt_matchinfo. */
static void print(const struct ipt_entry *entry, int numeric)
{
	printf("MAC ");
	print_mac(((struct ipt_mac_info *)&entry->matchinfo)->srcaddr,
		  ((struct ipt_mac_info *)&entry->matchinfo)->invert);
}

struct iptables_match mac
= { NULL,
    "mac",
    NETFILTER_VERSION,
    &help,
    &init,
    &parse,
    &final_check,
    &print,
    opts
};

void _init(void)
{
	register_match(&mac);
}
