/*
 * This is the 1999 rewrite of IP Firewalling, aiming for kernel 2.3.x.
 *
 * Copyright (C) 1999 Paul `Rusty' Russell & Michael J. Neuling
 */
#ifdef __KERNEL__
#include <linux/config.h>

#include <asm/smp.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/kmod.h>
#include <linux/stat.h>
#include <linux/vmalloc.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <linux/netdevice.h>
#else
#include "ip_tables_usersupp.h"
#endif

/* Don't need locking on lists: only looked at from userspace. */
#define ASSERT_READ_LOCK(x)
#define ASSERT_WRITE_LOCK(x)

#include "packet-filter/kernel/ip_tables.h"
#include "common/lockhelp.h"
#include "common/listhelp.h"

#ifndef IP_OFFSET
#define IP_OFFSET 0x1FFF
#endif

/* Understanding locking in this code: (thanks to Alan Cox for using
 * little words to explain this to me). -- PR
 *
 * In UP, there can be two packets traversing the chains:
 * 1) A packet from the current userspace context
 * 2) A packet off the bh handlers (timer or net).
 *
 * For SMP (kernel v2.1+), multiply this by # CPUs.
 *
 * [Note that this in not correct for 2.2 - because the socket code always
 *  uses lock_kernel() to serialize, and bottom halves (timers and net_bhs)
 *  only run on one CPU at a time.  This will probably change for 2.3.
 *  It is still good to use spinlocks because that avoids the global cli()
 *  for updating the tables, which is rather costly in SMP kernels -AK]
 *
 * This means counters and backchains can get corrupted if no precautions
 * are taken.
 *
 * To actually alter a chain on UP, we need only do a cli(), as this will
 * stop a bh handler firing, as we are in the current userspace context
 * (coming from a setsockopt()).
 *
 * On SMP, we need a write_lock_irqsave(), which is a simple cli() in
 * UP.
 *
 * For backchains and counters, we use an array, indexed by
 * [cpu_number_map[smp_processor_id()]*2 + !in_interrupt()]; the array
 * is of size [smp_num_cpus*2].  For v2.0, smp_num_cpus is effectively
 * 1.  So, confident of uniqueness, we modify counters even though we
 * only have a read lock (to read the counters, you need a write lock,
 * though).  */

/* Why I didn't use straight locking... -- PR
 *
 * The backchains can be separated out of the ip_chains structure, and
 * allocated as needed inside ipt_check().
 *
 * The counters, however, can't.  Trying to lock these means blocking
 * interrupts every time we want to access them.  This would suck HARD
 * performance-wise.  Not locking them leads to possible corruption,
 * made worse on 32-bit machines (counters are 64-bit).  */

/* A few tricks.  We avoid running off the end by keeping a special entry
 * there which uses ipt_error as its target function.  We avoid underflow
 * (more IPT_RETURN than jumps) by pointing the back pointer from the
 * entry points at this entry too.
 */

/*#define DEBUG_IP_FIREWALL*/
/*#define DEBUG_ALLOW_ALL*/ /* Useful for remote debugging */
/*#define DEBUG_IP_FIREWALL_USER*/
/*#define DEBUG_IP_FIREWALL_LOCKING*/

#ifdef CONFIG_SMP
#define SLOT_NUMBER() (cpu_number_map[smp_processor_id()]*2 + !in_interrupt())
#else
#define SLOT_NUMBER() (!in_interrupt())
#endif
#define NUM_SLOTS (smp_num_cpus*2)

#if defined(DEBUG_IP_FIREWALL) || defined (DEBUG_IP_FIREWALL_USER)
/* All the better to debug you with... */
#define static
#endif

#ifdef DEBUG_IP_FIREWALL_LOCKING
static unsigned int iptd_rlocks, iptd_wlocks;
#define IPTD_DEBUG_LOCK(d)			\
do {						\
	IPTD_DONT_HAVE_LOCK(d);			\
	d |= (1 << SLOT_NUMBER());		\
} while (0)

#define IPTD_DEBUG_UNLOCK(d)			\
do {						\
	IPTD_HAVE_LOCK(d);			\
	d &= ~(1 << SLOT_NUMBER());		\
} while (0)

#define IPTD_DONT_HAVE_LOCK(d)					\
do {								\
	if ((d) & (1 << SLOT_NUMBER()))				\
		printk("%s:%i: Got lock on %i already!\n", 	\
		       __FILE__, __LINE__, SLOT_NUMBER());	\
} while(0)

#define IPTD_HAVE_LOCK(d)				\
do {							\
	if (!((d) & (1 << SLOT_NUMBER())))		\
	printk("%s:%i:No lock on %i!\n", 		\
	       __FILE__, __LINE__, SLOT_NUMBER());	\
} while (0)

#else
#define IPTD_DEBUG_LOCK(d) do { } while(0)
#define IPTD_DEBUG_UNLOCK(d) do { } while(0)
#define IPTD_DONT_HAVE_LOCK(d) do { } while(0)
#define IPTD_HAVE_LOCK(d) do { } while(0)
#endif /*DEBUG_IP_FIREWALL_LOCKING*/

#define IPTD_READ_LOCK(l) do { IPTD_DEBUG_LOCK(iptd_rlocks); read_lock(l); } while (0)
#define IPTD_READ_UNLOCK(l) do { IPTD_DEBUG_UNLOCK(iptd_rlocks); read_unlock(l); } while (0)
#define IPTD_WRITE_LOCK_IRQ(l,f) do { IPTD_DEBUG_LOCK(iptd_wlocks); write_lock_irqsave(l,f); } while (0)
#define IPTD_WRITE_UNLOCK_IRQ(l,f) do { IPTD_DEBUG_UNLOCK(iptd_wlocks); write_unlock_irqrestore(l,f); } while (0)

/*
 *	Implement IP packet firewall
 */

#ifdef DEBUG_IP_FIREWALL
#define dprintf(format, args...)  printk(format , ## args)
#else
#define dprintf(format, args...)
#endif

#ifdef DEBUG_IP_FIREWALL_USER
#define duprintf(format, args...) printk(format , ## args)
#else
#define duprintf(format, args...)
#endif

/* Lock around ipt_table_info table */
rwlock_t ipt_lock = RW_LOCK_UNLOCKED;

struct ipt_kern_entry
{
	struct ipt_match *match;
	struct ipt_ip ip;
	union ipt_matchinfo matchinfo;
	struct ipt_target *act;
	unsigned int cacheinfo;
	union ipt_targinfo targinfo;
};

/* The table itself */
struct ipt_table_info
{
	unsigned int number;
	unsigned int hook_entry[3];
	unsigned int underflow[3];

	/* Pointer to NUM_SLOTS() * number entries each. */
	struct ipt_counters *counters;
	unsigned int *back;

	/* Actually number of these */
	struct ipt_kern_entry table[0];
};

static struct ipt_table_info *ipt_table_info;
static LIST_HEAD(ipt_target);
static LIST_HEAD(ipt_match);

#define ADD_COUNTER(c,b,p) do { (c).bcnt += (b); (c).pcnt += (p); } while(0)

/* Returns whether matches rule or not. */
static int
ip_packet_match(const struct iphdr *ip,
		const char *indev,
		const char *outdev,
		const struct ipt_ip *ipinfo,
		int isfrag)
{
	size_t i;
	unsigned long ret;

#define FWINV(bool,invflg) ((bool) ^ !!(ipinfo->invflags & invflg))
	/*
	 *	This is a bit simpler as we don't have to walk
	 *	an interface chain as you do in BSD - same logic
	 *	however.
	 */

	if (FWINV((ip->saddr&ipinfo->smsk.s_addr) != ipinfo->src.s_addr,
		  IPT_INV_SRCIP)
	    || FWINV((ip->daddr&ipinfo->dmsk.s_addr) != ipinfo->dst.s_addr,
		     IPT_INV_DSTIP)) {
		dprintf("Source or dest mismatch.\n");

		dprintf("SRC: %u. Mask: %u. Target: %u.%s\n", ip->saddr,
			ipinfo->smsk.s_addr, ipinfo->src.s_addr,
			ipinfo->invflags & IPT_INV_SRCIP ? " (INV)" : "");
		dprintf("DST: %u. Mask: %u. Target: %u.%s\n", ip->daddr,
			ipinfo->dmsk.s_addr, ipinfo->dst.s_addr,
			ipinfo->invflags & IPT_INV_DSTIP ? " (INV)" : "");
		return 0;
	}

	/* Look for ifname matches; this should unroll nicely. */
	for (i = 0, ret = 0; i < IFNAMSIZ/sizeof(unsigned long); i++) {
		ret |= (((const unsigned long *)indev)[i]
			^ ((const unsigned long *)ipinfo->iniface)[i])
			& ((const unsigned long *)ipinfo->iniface_mask)[i];
	}

	if (FWINV(ret != 0, IPT_INV_VIA_IN)) {
		dprintf("VIA in mismatch (%s vs %s).%s\n",
			indev, ipinfo->iniface,
			ipinfo->invflags&IPT_INV_VIA_IN ?" (INV)":"");
		return 0;
	}

	for (i = 0, ret = 0; i < IFNAMSIZ/sizeof(unsigned long); i++) {
		ret |= (((const unsigned long *)outdev)[i]
			^ ((const unsigned long *)ipinfo->outiface)[i])
			& ((const unsigned long *)ipinfo->outiface_mask)[i];
	}

	if (FWINV(ret != 0, IPT_INV_VIA_OUT)) {
		dprintf("VIA out mismatch (%s vs %s).%s\n",
			outdev, ipinfo->outiface,
			ipinfo->invflags&IPT_INV_VIA_OUT ?" (INV)":"");
		return 0;
	}

	/* If we have a fragment rule but the packet is not a fragment
	 * the we return zero */
	if (FWINV((ipinfo->flags&IPT_F_FRAG) && !isfrag, IPT_INV_FRAG)) {
		dprintf("Fragment rule but not fragment.%s\n",
			ipinfo->invflags & IPT_INV_FRAG ? " (INV)" : "");
		return 0;
	}

	/* Check TOS match */
	if ((ipinfo->flags & IPT_F_TOS)
	    && FWINV(ipinfo->tos != ip->tos, IPT_INV_TOS)) {
		dprintf("Packet TOS %u does not match %u.%s\n",
			ip->tos, ipinfo->tos,
			ipinfo->invflags&IPT_INV_TOS ? " (INV)":"");
		return 0;
	}

	/* Check specific protocol */
	if (ipinfo->proto
	    && FWINV(ip->protocol != ipinfo->proto, IPT_INV_PROTO)) {
		dprintf("Packet protocol %hi does not match %hi.%s\n",
			ip->protocol, ipinfo->proto,
			ipinfo->invflags&IPT_INV_PROTO ? " (INV)":"");
		return 0;
	}

	return 1;
}

static int ip_checkentry(const struct ipt_ip *ip)
{
	if (ip->flags & ~IPT_F_MASK) {
		duprintf("Unknown flag bits set: %08X\n",
			 ip->flags & ~IPT_F_MASK);
		return 0;
	}
	if (ip->invflags & ~IPT_INV_MASK) {
		duprintf("Unknown invflag bits set: %08X\n",
			 ip->invflags & ~IPT_INV_MASK);
		return 0;
	}
	return 1;
}

static int
ipt_standard_target(struct sk_buff *skb,
		    const char *indev,
		    const char *outdev,
		    const union ipt_targinfo *info,
		    unsigned int pos)
{
	return info->standard.verdict;
}

static int
ipt_error(struct sk_buff *skb,
	  const char *indev,
	  const char *outdev,
	  const union ipt_targinfo *info,
	  unsigned int pos)
{
	if (net_ratelimit())
		printk("ip_tables: error: `%s'\n", info->error);

	return -NF_DROP - 1;
}

/*
 * Returns one of the generic firewall policies, like FW_ACCEPT.  */
static unsigned int
ipt_check(struct sk_buff *skb,
	  unsigned int pos,
	  unsigned int underflow,
	  const char *indev,
	  const char *outdev,
	  unsigned int *cache)
{
	u_int16_t offset;
	struct iphdr *ip = skb->nh.iph;
	void *protohdr = (u_int32_t *)ip + ip->ihl;
	u_int16_t datalen = skb->len - ip->ihl * 4;
	unsigned int level = 1;
	const unsigned int number = ipt_table_info->number;
	unsigned int *stack = &ipt_table_info->back[SLOT_NUMBER() * number];
	int spos = (int)pos;
	int hotdrop = 0;
	struct ipt_counters *counters
		= &ipt_table_info->counters[SLOT_NUMBER() * number];

	/* Prime the stack for return from builtin chain */
	stack[0] = underflow;

	/* We handle fragments by dealing with the first fragment as
	 * if it was a normal packet.  All other fragments are treated
	 * normally, except that they will NEVER match rules that ask
	 * things we don't know, ie. tcp syn flag or ports).  If the
	 * rule is also a fragment-specific rule, non-fragments won't
	 * match it. */
	offset = ntohs(ip->frag_off) & IP_OFFSET;

	while (!hotdrop) {
		unsigned int oldpos = (unsigned int)spos;
		struct ipt_kern_entry *e = &ipt_table_info->table[oldpos];
		*cache |= e->cacheinfo;
		if (ip_packet_match(skb->nh.iph, indev, outdev, &e->ip,
				    offset)
		    && (!e->match->match
			|| e->match->match(skb, indev, outdev, &e->matchinfo,
					   offset, protohdr, datalen,
					   &hotdrop))) {
			dprintf("We have a match on %u.\n", oldpos);
			ADD_COUNTER(counters[oldpos], ntohs(ip->tot_len), 1);

			spos = e->act->target(skb, indev, outdev, &e->targinfo,
					      spos);

			/* Pop from stack? */
			if (spos == -IPT_RETURN - 1) {
				dprintf("ipt_check: pop %i to %u.\n",
					oldpos, stack[level-1]);
				spos = stack[--level];
			}
			/* A verdict. */
			else if (spos < 0) break;
			/* Push onto stack? */
			else if (spos != (int)oldpos+1) {
				dprintf("ipt_check: backptr %i to %u.\n",
					spos, oldpos);
				stack[level++] = oldpos+1;
			}
		}
		else spos++;
	}

	dprintf("Answer is %i\n", -spos - 1);
#ifdef DEBUG_ALLOW_ALL
	return NF_ACCEPT;
#else
	if (hotdrop)
		return NF_DROP;
	else return (unsigned int)(-spos - 1);
#endif
}

static void *
find_inlist(struct list_head *head, const char *name)
{
	void *ret;

	duprintf("find_inlist: searching for `%s' in %s.\n",
		 name, head == &ipt_target ? "ipt_target"
		 : head == &ipt_match ? "ipt_match" : "UNKNOWN");

	ret = list_named_find(head, name);
#ifndef ALWAYS_IN_INTERRUPT
	/* Autoloading not allowed if in interrupt. */
	if (!ret) {
		char modulename[IPT_FUNCTION_MAXNAMELEN + sizeof("ipt_")];
		strcpy(modulename, "ipt_");
		strcat(modulename, name);
		duprintf("find_inlist: loading `%s'.\n", modulename);
		request_module(modulename);
		ret = list_named_find(head, name);
	}
#endif
	return ret;
}

/* All zeroes == unconditional rule */
static inline int
unconditional(const struct ipt_ip *ip)
{
	unsigned int i;

	for (i = 0; i < sizeof(*ip); i++)
		if (((char *)ip)[i])
			break;

	return (i == sizeof(*ip));
}

/* Figures out from what chain each rule can be called: returns 0 if
   there are loops.  Puts chain start bitmask in back[]. */
static int
mark_source_chains(struct ipt_table_info *newinfo)
{
	unsigned int chain;
	unsigned int *comefrom = newinfo->back + newinfo->number;

	/* No recursion; use newinfo->back to save source chains,
	   and newinfo->back + number of rules to save back ptrs. */
	memset(newinfo->back, 0, sizeof(unsigned int)*newinfo->number);

	for (chain = 0; chain < 3; chain++) {
		unsigned int rule = newinfo->hook_entry[chain];

		/* Set initial back pointer. */
		comefrom[rule] = rule;

		for (;;) {
			duprintf("Examining rule %u\n", rule);
			if (newinfo->back[rule] & (1 << 3)) {
				dprintf("Loop in chain %s at rule %u.\n",
					chain == IPT_HOOK_INPUT ? "INPUT"
					: chain == IPT_HOOK_FORWARD ? "FORWARD"
					: chain == IPT_HOOK_OUTPUT ? "OUTPUT"
					: "screwup", rule);
				return 0;
			}
			newinfo->back[rule] |= ((1 << chain) | (1 << 3));

			/* Unconditional return/END. */
			if (!newinfo->table[rule].match->match
			    && (newinfo->table[rule].act->target
				== ipt_standard_target)
			    && (newinfo->table[rule].targinfo.standard.verdict
				< 0)
			    && unconditional(&newinfo->table[rule].ip)) {
				unsigned int oldrule;
				duprintf("Return at rule %u\n", rule);
				/* Return: backtrack through the last
                                   big jump. */
				do {
					newinfo->back[rule] ^= (1 << 3);
#ifdef DEBUG_IP_FIREWALL_USER
					if (newinfo->back[rule] & (1 << 3)) {
						duprintf("Back unset on"
							 " ch %u rule %u\n",
							 chain, rule);
					}
#endif
					oldrule = rule;
					rule = comefrom[rule];

					/* We're at the start. */
					if (rule == oldrule)
						goto next;
				} while (rule == oldrule-1);

				/* Move along one */
				comefrom[rule+1] = rule;
				rule++;
			} else {
				int newrule = (newinfo->table[rule].targinfo
					       .standard.verdict);

				if (newinfo->table[rule].act->target
				    == ipt_standard_target
				    && newrule >= 0) {
					/* This a jump; chase it. */
					duprintf("Jump rule %u -> %u\n",
						 rule, newrule);
				} else {
					duprintf("Fall thru rule %u\n", rule);
					/* ... this is a fallthru */
					newrule = rule+1;
				}
				comefrom[newrule] = rule;
				rule = newrule;
			}
		}
		next:
		duprintf("Finished chain %u\n", chain);
	}
	return 1;
}

/* Layout of table in memory:

 ipt_table_info ->  struct ipt_table_info
                    struct ipt_kern_entry [number of rules]
		
                    struct ipt_counters [NUM_SLOTS * number of rules]
                    unsigned int stack [NUM_SLOTS * number of rules]
*/

/* Checks the user-supplied table segment. */
static int
translate_table(struct ipt_table_info *newinfo,
		const struct ipt_replace *repl)
{
	unsigned int i;
	struct ipt_counters *counters = (void *)&newinfo->table[repl->number];
	unsigned int *stack = (void *)&counters[NUM_SLOTS * repl->number];

	/* Check hooks. */
	for (i = 0; i < sizeof(repl->hook_entry)/sizeof(unsigned int); i++) {
		if (repl->hook_entry[i] >= repl->number) {
			duprintf("Bad hook %u->%u.\n", i, repl->hook_entry[i]);
			return -EINVAL;
		}
		if (repl->underflow[i] >= repl->number) {
			duprintf("Bad underflow %u->%u.\n", i,
				 repl->hook_entry[i]);
			return -EINVAL;
		}
	}

	duprintf("translate_table: hooks OK.\n");
	*newinfo = ((struct ipt_table_info)
		    { repl->number,
		      { repl->hook_entry[0],
			repl->hook_entry[1],
			repl->hook_entry[2] },
		      { repl->underflow[0],
			repl->underflow[1],
			repl->underflow[2] },
		      counters, stack,
		      { } });

	/* Zero counters */
	memset(counters, 0,
	       sizeof(struct ipt_counters) * NUM_SLOTS * repl->number);

	/* Each target name must be valid. */
	for (i = 0; i < repl->number; i++) {
		const struct ipt_entry *e = &repl->entry_array[i];
		struct ipt_target *act;
		struct ipt_match *match;

		act = find_inlist(&ipt_target, e->target_name);
		if (!act) {
			duprintf("ip_tables: no target by name `%s'.\n",
				 e->target_name);
			return -EINVAL;
		}

		match = find_inlist(&ipt_match, e->match_name);
		if (!match) {
			duprintf("ip_tables: no match by name `%s'.\n",
				 e->match_name);
			return -EINVAL;
		}

		newinfo->table[i]
			= ((struct ipt_kern_entry)
			   { match, e->ip, e->matchinfo, act, e->nfcache,
			     e->targinfo });

		/* FIXME: Do we need this? --RR */
		/* Set counters in the first slot. */
		newinfo->counters[i] = e->counters;
	}

	if (!mark_source_chains(newinfo))
		return -ELOOP;

	/* Each sanity check must pass */
	for (i = 0; i < repl->number; i++) {
		struct ipt_kern_entry *k = &newinfo->table[i];
		if (k->act->checkentry
		    && !k->act->checkentry(&k->targinfo, i, newinfo->number,
					   newinfo->back[i])) {
			duprintf("ip_tables: check failed for `%s'.\n",
				 k->act->name);
			return -EINVAL;
		}
		if ((k->match->checkentry
		     && !k->match->checkentry(&k->ip, &k->matchinfo,
					      newinfo->back[i]))
		    || !ip_checkentry(&k->ip)) {
			duprintf("ip_tables: check failed for `%s'.\n",
				 k->match->name);
			return -EINVAL;
		}
	}

	return 0;
}

static void
sum_counters(struct ipt_counters *total,
	     const struct ipt_counters *counters,
	     unsigned int offset,
	     unsigned int number)
{
	unsigned int i;

	total->pcnt = total->bcnt = 0;
	for (i = 0; i < NUM_SLOTS; i++) {
		ADD_COUNTER(*total,
			    counters[number * i + offset].bcnt,
			    counters[number * i + offset].pcnt);
	}
}

/* Gets counters. */
static void
get_counters(const struct ipt_table_info *info,
	     struct ipt_counters counters[])
{
	unsigned int i;

#ifdef DEBUG_IP_FIREWALL_LOCKING
	/* Need lock if operating on live data */
	if (info == ipt_table_info)
		IPTD_HAVE_LOCK(iptd_wlocks);
#endif

	for (i = 0; i < info->number; i++)
		sum_counters(&counters[i], info->counters, i, info->number);
}

static int
get_entries(struct ipt_get_entries *entries)
{
	unsigned long flags;
	unsigned int i;

	IPTD_WRITE_LOCK_IRQ(&ipt_lock, flags);
	if (entries->num_entries != ipt_table_info->number) {
		IPTD_WRITE_UNLOCK_IRQ(&ipt_lock, flags);
		duprintf("get_entries: I've got %u not %u!\n",
			 ipt_table_info->number, entries->num_entries);
		return -EINVAL;
	}
	for (i = 0; i < ipt_table_info->number; i++) {
		/* Translate back to userland friendly. */
		struct ipt_kern_entry *kent = &ipt_table_info->table[i];
		struct ipt_entry *ent = &entries->entries[i];

		ent->ip = kent->ip;
		ent->matchinfo = kent->matchinfo;
		ent->nfcache = kent->cacheinfo;
		memcpy(ent->target_name, kent->act->name,
		       sizeof(ent->target_name));
		memcpy(ent->match_name, kent->match->name,
		       sizeof(ent->match_name));
		duprintf("get_entries: copying target `%s' match `%s'\n",
			 ent->target_name, ent->match_name);
		sum_counters(&ent->counters, ipt_table_info->counters, i,
			     ipt_table_info->number);
		ent->targinfo = kent->targinfo;
	}
	IPTD_WRITE_UNLOCK_IRQ(&ipt_lock, flags);
	return 0;
}

static int
add_counters(struct ipt_counters_info *paddc)
{
	unsigned long flags;
	unsigned int i;

	IPTD_WRITE_LOCK_IRQ(&ipt_lock, flags);
	if (ipt_table_info->number != paddc->num_counters) {
		IPTD_WRITE_UNLOCK_IRQ(&ipt_lock, flags);
		return -EINVAL;
	}
	for (i = 0; i < ipt_table_info->number; i++) {
		/* We're lazy, and add to the first slot; overflow
		 * works its fey magic and everything is OK. */
		ADD_COUNTER(ipt_table_info->counters[i],
			    paddc->counters[i].bcnt,
			    paddc->counters[i].pcnt);
	}
	IPTD_WRITE_UNLOCK_IRQ(&ipt_lock, flags);
	return 0;
}

static struct ipt_table_info *
replace_table(struct ipt_replace *repl, int *error)
{
	unsigned int i;
	struct ipt_table_info *newinfo, *oldinfo;
	unsigned long flags;

	/* Allocate slots and new entries: alignment should be OK */
	newinfo = vmalloc(sizeof(struct ipt_table_info)
			  + repl->number * sizeof(struct ipt_kern_entry)
			  + (repl->number * NUM_SLOTS
			     * sizeof(struct ipt_counters))
			  + (repl->number * NUM_SLOTS
			     * sizeof(unsigned int)));
	if (!newinfo) {
		*error = -ENOMEM;
		return NULL;
	}

	*error = translate_table(newinfo, repl);
	if (*error)
		goto bad_data;

	duprintf("ip_tables: Translated table\n");
	/* Increase module usage counts. */
	for (i = 0; i < newinfo->number; i++) {
		if (newinfo->table[i].act->me)
			__MOD_INC_USE_COUNT(newinfo->table[i].act->me);
		if (newinfo->table[i].match->me)
			__MOD_INC_USE_COUNT(newinfo->table[i].match->me);
	}

	/* Do the substitution. */
	IPTD_WRITE_LOCK_IRQ(&ipt_lock, flags);
	/* Check inside lock: is the number of counters correct? */
	if (repl->num_counters != ipt_table_info->number) {
		IPTD_WRITE_UNLOCK_IRQ(&ipt_lock, flags);
		*error = -EINVAL;
		goto bad_data;
	}
	oldinfo = ipt_table_info;
	ipt_table_info = newinfo;
	IPTD_WRITE_UNLOCK_IRQ(&ipt_lock, flags);

	return oldinfo;

 bad_data:
	vfree(newinfo);
	return NULL;
}

static void *
alloc_from_user(size_t supposed_len_base,
			     size_t size_per_entry,
			     void *user, unsigned int len,
			     int *error)
{
	unsigned int number;
	void *ret;

	if ((*error = get_user(number, (unsigned int *)user)) != 0)
		return NULL;

	if (supposed_len_base + number * size_per_entry != len) {
		duprintf("ip_tables: bad length (%u = %u + %u * %u not %u).\n",
			 supposed_len_base + number * size_per_entry,
			 supposed_len_base, number, size_per_entry, len);
		*error = -EINVAL;
		return NULL;
	}
	ret = vmalloc(len);
	if (!ret) {
		duprintf("ip_tables: oom (%u).\n", len);
		*error = -ENOMEM;
	}
	*((unsigned int *)ret) = number;
	return ret;
}

static void *
var_from_user(size_t supposed_len_base,
			   size_t size_per_entry,
			   void *user, unsigned int len,
			   int *error)
{
	void *ret;

	if ((ret = alloc_from_user(supposed_len_base, size_per_entry, user,
				   len, error)) != NULL) {
		if (copy_from_user(ret, user, len) != 0) {
			*error = -EFAULT;
			vfree(ret);
			ret = NULL;
		}
	}
	return ret;
}

static void
cleanup_entry(struct ipt_kern_entry *ent)
{
	if (ent->act->deleteentry)
		ent->act->deleteentry(&ent->targinfo);
	if (ent->match->deleteentry)
		ent->match->deleteentry(&ent->matchinfo);
	if (ent->act->me)
		__MOD_DEC_USE_COUNT(ent->act->me);
	if (ent->match->me)
		__MOD_DEC_USE_COUNT(ent->match->me);
}

static int
do_replace(void *user, unsigned int len)
{
	int ret = 0;
	struct ipt_replace *repl;

	repl = var_from_user(sizeof(struct ipt_replace),
			     sizeof(struct ipt_entry),
			     user, len, &ret);
	if (repl) {
		struct ipt_counters *uout = repl->counters;
		repl->counters
			= vmalloc(repl->num_counters
				  * sizeof(struct ipt_counters));
		if (!repl->counters)
			ret = -ENOMEM;
		else {
			struct ipt_table_info *oldinfo;

			oldinfo = replace_table(repl, &ret);
			if (oldinfo) {
				unsigned int i;
				/* Get the old counters. */
				get_counters(oldinfo, repl->counters);

				/* Decrease module usage counts
				   and free resource */
				for (i = 0; i < oldinfo->number; i++)
					cleanup_entry(&oldinfo->table[i]);

				vfree(oldinfo);
				/* Silent error: too late now. */
				copy_to_user(uout, repl->counters,
					     sizeof(struct ipt_counters)
					     * repl->num_counters);
			}
			vfree(repl->counters);
		}
		vfree(repl);
	}
	return ret;
}

static int
do_ipt_set_ctl(struct sock *sk,	int cmd, void *user, unsigned int len)
{
	int ret;

	if (!capable(CAP_NET_ADMIN))
		return -EPERM;

	switch (cmd) {
	case IPT_SO_SET_REPLACE:
		ret = do_replace(user, len);
		break;

	case IPT_SO_SET_ADD_COUNTERS: {
		struct ipt_counters_info *pgetc
			= var_from_user(sizeof(struct ipt_counters_info),
					sizeof(struct ipt_counters),
					user, len, &ret);
		if (pgetc) {
			ret = add_counters(pgetc);
			vfree(pgetc);
		}
		break;
	}

	default:
		duprintf("do_ipt_set_ctl:  unknown request %i\n", cmd);
		ret = -EINVAL;
	}

	return ret;
}

static int
do_ipt_get_ctl(struct sock *sk, int cmd, void *user, int *len)
{
	int ret;

	if (!capable(CAP_NET_ADMIN))
		return -EPERM;

	switch (cmd) {
	case IPT_SO_GET_INFO:
		if (*len != sizeof(struct ipt_getinfo))
			ret = -EINVAL;
		else {
			struct ipt_getinfo info
				= { { ipt_table_info->hook_entry[0],
				      ipt_table_info->hook_entry[1],
				      ipt_table_info->hook_entry[2] },
				    { ipt_table_info->underflow[0],
				      ipt_table_info->underflow[1],
				      ipt_table_info->underflow[2] },
				    ipt_table_info->number };
			if (copy_to_user(user, &info, *len) != 0)
				ret = -EFAULT;
			else
				ret = 0;
		}
		break;

	case IPT_SO_GET_ENTRIES: {
		struct ipt_get_entries *entries
			= alloc_from_user(sizeof(struct ipt_get_entries),
					  sizeof(struct ipt_entry),
					  user, *len, &ret);
		if (entries) {
			if ((ret = get_entries(entries)) == 0) {
				if (copy_to_user(user, entries,
						 ipt_table_info->number
						 * sizeof(struct ipt_entry))
				    != 0)
					ret = -EFAULT;
			}
			vfree(entries);
		}
		break;
	}

	case IPT_SO_GET_VERSION:
		if (*len < sizeof(NETFILTER_VERSION)+1)
			ret = -EINVAL;
		else if (copy_to_user(user, NETFILTER_VERSION,
				      sizeof(NETFILTER_VERSION)+1) != 0)
			ret = -EFAULT;
		else {
			*len = sizeof(NETFILTER_VERSION)+1;
			ret = 0;
		}
		break;
		
	default:
		duprintf("do_ipt_get_ctl:  unknown request %i\n", cmd);
		ret = -EINVAL;
	}

	return ret;
}

/* The work comes in here from netfilter.c. */
static unsigned int
ipt_hook(unsigned int hook,
	 struct sk_buff **pskb,
	 const struct net_device *indev,
	 const struct net_device *outdev)
{
	unsigned int ret;

	dprintf("ipt_hook: hook %u.\n", hook);
	IPTD_READ_LOCK(&ipt_lock);
	ret = ipt_check(*pskb, ipt_table_info->hook_entry[hook-NF_IP_LOCAL_IN],
			ipt_table_info->underflow[hook-NF_IP_LOCAL_IN],
			indev ? indev->name : "",
			outdev ? outdev->name : "",
			&(*pskb)->nfcache);
	IPTD_READ_UNLOCK(&ipt_lock);
	dprintf("ipt_hook: returning %u.\n", ret);

	return ret;
}

static unsigned int
ipt_local_out_hook(unsigned int hook,
		   struct sk_buff **pskb,
		   const struct net_device *indev,
		   const struct net_device *outdev)
{
	/* root is playing with raw sockets. */
	if ((*pskb)->len < sizeof(struct iphdr)
	    || (*pskb)->nh.iph->ihl * 4 < sizeof(struct iphdr)) {
		dprintf("ipt_hook: happy cracking.\n");
		return NF_ACCEPT;
	}

	return ipt_hook(hook, pskb, indev, outdev);
}

/* Registration hooks for targets. */
int
ipt_register_target(struct ipt_target *target)
{
	if (strcmp(target->version, NETFILTER_VERSION) != 0) {
		printk("iptables: target %s is v%s (I'm version %s)\n",
		       target->name, target->version, NETFILTER_VERSION);
		return -EINVAL;
	}

	if (list_named_insert(&ipt_target, target)) {
		MOD_INC_USE_COUNT;
		return 0;
	}

	duprintf("ipt_register_target: `%s' already in list!\n",
		 target->name);
	return -EINVAL;
}

void
ipt_unregister_target(struct ipt_target *target)
{
	LIST_DELETE(&ipt_target, target);
	MOD_DEC_USE_COUNT;
}

int
ipt_register_match(struct ipt_match *match)
{
	if (strcmp(match->version, NETFILTER_VERSION) != 0) {
		printk("iptables: match %s is v%s (I'm version %s)\n",
		       match->name, match->version, NETFILTER_VERSION);
		return -EINVAL;
	}

	if (list_named_insert(&ipt_match, match)) {
		MOD_INC_USE_COUNT;
		return 0;
	}

	duprintf("ipt_register_match: `%s' already in list!\n",
		 match->name);
	return -EINVAL;
}

void
ipt_unregister_match(struct ipt_match *match)
{
	LIST_DELETE(&ipt_match, match);
	MOD_DEC_USE_COUNT;
}

static int
ipt_standard_checkentry(const union ipt_targinfo *info,
			unsigned int pos,
			unsigned int finalpos,
			unsigned int chain_mask)
{
	/* Check standard info. */
	if (info->standard.verdict < -IPT_RETURN - 1
	    || info->standard.verdict > (int)finalpos) {
		duprintf("ipt_standard_check: bad verdict (%i)\n",
			 info->standard.verdict);
		return 0;
	}
	return 1;
}

static struct nf_hook_ops ipt_ops[]
= { { { NULL, NULL }, ipt_hook, NULL, PF_INET, NF_IP_LOCAL_IN, 0 },
    { { NULL, NULL }, ipt_hook, NULL, PF_INET, NF_IP_FORWARD, 0 },
    { { NULL, NULL }, ipt_local_out_hook, NULL, PF_INET, NF_IP_LOCAL_OUT, 0 }
};

/* The built-in targets: error and standard. */
static struct ipt_target ipt_error_reg
= { { NULL, NULL }, IPT_ERROR_TARGET, NETFILTER_VERSION,
    ipt_error, NULL, NULL, NULL };
static struct ipt_target ipt_standard_reg
= { { NULL, NULL }, IPT_STANDARD_TARGET, NETFILTER_VERSION,
    ipt_standard_target, ipt_standard_checkentry, NULL, NULL };

/* The built-in match: normal IP match. */
static struct ipt_match ipt_standard_match
= { { NULL, NULL }, IPT_STANDARD_MATCH, NULL, NULL, NULL };
static struct nf_sockopt_ops ipt_sockopts
= { { NULL, NULL }, PF_INET, IPT_BASE_CTL, IPT_SO_SET_MAX+1, do_ipt_set_ctl,
    IPT_BASE_CTL, IPT_SO_GET_MAX+1, do_ipt_get_ctl };

/* ACCEPT rule, forward rule (if module, verdict may be modified on
   module load), ACCEPT rule, and an error rule. */

/* Looking for memcpy?  If this struct changes, could get called here. */
static struct
{
	struct ipt_replace repl;
	struct ipt_entry entries[4];
} initial_table __initdata
= { { 4, { 0, 1, 2 }, { 0, 1, 2}, 0, NULL, { } },
    { { { { 0 }, { 0 }, { 0 }, { 0 }, "", "", { 0 }, { 0 }, 0, 0 },
	{ { { 0, 0 }, { 0, 0 }, 0, 0, 0 } },
	0, "", "", { 0, 0 },
	{ { -NF_ACCEPT - 1, { 0, 0, 0 } } } },
      { { { 0 }, { 0 }, { 0 }, { 0 }, "", "", { 0 }, { 0 }, 0, 0 },
	{ { { 0, 0 }, { 0, 0 }, 0, 0, 0 } },
	0, "", "", { 0, 0 },
	{ { -NF_ACCEPT - 1, { 0, 0, 0 } } } },
      { { { 0 }, { 0 }, { 0 }, { 0 }, "", "", { 0 }, { 0 }, 0, 0 },
	{ { { 0, 0 }, { 0, 0 }, 0, 0, 0 } },
	0, "", "", { 0, 0 },
	{ { -NF_ACCEPT - 1, { 0, 0, 0 } } } },
      { { { 0 }, { 0 }, { 0 }, { 0 }, "", "", { 0 }, { 0 }, 0, 0 },
	{ { { 0, 0 }, { 0, 0 }, 0, 0, 0 } },
	0, "", IPT_ERROR_TARGET, { 0, 0 },
	{ { 0, { 0, 0, 0 } } } }
    }
};

/* Default to no forward for security reasons. */
static int forward = NF_DROP;
MODULE_PARM(forward, "i");

/* FIXME: Panic if this fails. */
int __init init(void)
{
	unsigned int i;
	struct ipt_table_info bootstrap
		= { 0, { 0, 0, 0 }, { 0, 0, 0 }, NULL, NULL, { } };
	int ret;

#ifdef DEBUG_IP_FIREWALL_LOCKING
	iptd_wlocks = iptd_rlocks = 0;
#endif

	if (forward < 0 || forward > NF_MAX_VERDICT) {
		printk("iptables v%s forward must be 0 or 1\n",
		       NETFILTER_VERSION);
		return -EINVAL;
	}

	initial_table.entries[IPT_HOOK_FORWARD].targinfo.standard.verdict
		= -forward - 1;

	list_prepend(&ipt_match, &ipt_standard_match);
	list_append(&ipt_target, &ipt_standard_reg);
	list_append(&ipt_target, &ipt_error_reg);

	/* Ensure counter counts match. */
	ipt_table_info = &bootstrap;

	replace_table(&initial_table.repl, &ret);
	if (ret < 0) {
		duprintf("Unable to set initial table.\n");
		return ret;
	}

	/* Register setsockopt */
	ret = nf_register_sockopt(&ipt_sockopts);
	if (ret < 0) {
		duprintf("Unable to register sockopts.\n");
		return ret;
	}

	/* Register hooks */
	for (i = 0; i < sizeof(ipt_ops)/sizeof(struct nf_hook_ops); i++) {
		ret = nf_register_hook(&ipt_ops[i]);
		if (ret < 0) {
			duprintf("Unable to register iptable on hook %u.\n",
				 i+NF_IP_LOCAL_IN);
			return ret;
		}
	}

#if defined(DEBUG_IP_FIREWALL) || defined(DEBUG_IP_FIREWALL_USER)
	printk("Firewall tables: ``So... how is this better than KDE?''.\n");
#endif
	return 0;
}

void __exit cleanup(void)
{
	unsigned long flags;
	unsigned int i;

	IPTD_WRITE_LOCK_IRQ(&ipt_lock, flags);

	for (i = 0; i < sizeof(ipt_ops)/sizeof(struct nf_hook_ops); i++)
		nf_unregister_hook(&ipt_ops[i]);
	nf_unregister_sockopt(&ipt_sockopts);
	vfree(ipt_table_info);

	IPTD_WRITE_UNLOCK_IRQ(&ipt_lock, flags);
}

module_init(init);
module_exit(cleanup);
