/* Simple program to turn ip packet into an icmp error.  */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <netdb.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>
#if defined(__GLIBC__) && __GLIBC__ == 2
#include <net/ethernet.h>
#else
#include <linux/if_ether.h>
#endif
#include <sys/types.h>

static u_int16_t
csum_partial(void *buffer, unsigned int len, u_int16_t prevsum)
{
	u_int32_t sum = 0;
	u_int16_t *ptr = buffer;

	while (len > 1)  {
		sum += *ptr++;
		len -= 2;
	}
	if (len) {
		union {
			u_int8_t byte;
			u_int16_t wyde;
		} odd;
		odd.wyde = 0;
		odd.byte = *((u_int8_t *)ptr);
		sum += odd.wyde;
	}
	sum = (sum >> 16) + (sum & 0xFFFF);
	sum += prevsum;
	return (sum + (sum >> 16));
}

struct packet {
	u_int16_t pad;
#if defined(__GLIBC__) && __GLIBC__ == 2
	struct ether_header ehdr;
#else
	struct ethhdr ehdr;
#endif
	struct iphdr iph;
	union {
		struct tcphdr tcph;
		struct udphdr udph;
		struct icmphdr icmph;
		unsigned char raw[0];
	} u;
	unsigned char data[65535];
};

static int
string_to_number(const char *s, int min, int max)
{
	int number;
	char *end;

	number = (int)strtol(s, &end, 10);
	if (*end == '\0' && end != s) {
		/* we parsed a number, let's see if we want this */
		if (min <= number && number <= max)
			return number;
		else
			return -1;
	} else
		return -1;
}

static struct in_addr *
dotted_to_addr(const char *dotted)
{
	static struct in_addr addr;
	unsigned char *addrp;
	char *p, *q;
	int onebyte, i;
	char buf[20];

	/* copy dotted string, because we need to modify it */
	strncpy(buf, dotted, sizeof(buf) - 1);
	addrp = (unsigned char *) &(addr.s_addr);

	p = buf;
	for (i = 0; i < 3; i++) {
		if ((q = strchr(p, '.')) == NULL)
			return (struct in_addr *) NULL;
		else {
			*q = '\0';
			if ((onebyte = string_to_number(p, 0, 255)) == -1)
				return (struct in_addr *) NULL;
			else
				addrp[i] = (unsigned char) onebyte;
		}
		p = q + 1;
	}

	/* we've checked 3 bytes, now we check the last one */
	if ((onebyte = string_to_number(p, 0, 255)) == -1)
		return (struct in_addr *) NULL;
	else
		addrp[3] = (unsigned char) onebyte;

	return &addr;
}

static void
create_icmp(struct packet *packet, const struct packet *orig_packet,
	    const char *srcip, const char *typestr, const char *codestr)
{
	const struct in_addr *addr;
	int type, code;
	/* Length = IP header + ICMP header + Orig IP header length + 8 */
	u_int16_t len = sizeof(struct iphdr) + sizeof(struct icmphdr)
		+ orig_packet->iph.ihl * 4 + 8;

	addr = dotted_to_addr(srcip);
	if (!addr) {
		fprintf(stderr, "Bad src address `%s'\n", srcip);
		exit(1);
	}

	packet->iph = ((struct iphdr)
		      {
#if __BYTE_ORDER == __LITTLE_ENDIAN
			      5, 4,
#else
			      4, 5,
#endif
			      0, htons(len), 0, 0, 255, IPPROTO_ICMP, 0, addr->s_addr,
			      orig_packet->iph.saddr });
	packet->iph.check = ~csum_partial(&packet->iph, sizeof(packet->iph), 0);

	if ((type = string_to_number(typestr, 0, 255)) < 0) {
		fprintf(stderr, "Bad icmp type: `%s'\n", typestr);
		exit(1);
	}

	if ((code = string_to_number(codestr, 0, 255)) < 0) {
		fprintf(stderr, "Bad icmp code: `%s'\n", codestr);
		exit(1);
	}

	packet->u.icmph.type = type;
	packet->u.icmph.code = code;
	packet->u.icmph.checksum = 0;
	packet->u.icmph.un.gateway = 0; /* MBZ */

	memcpy(packet->u.raw + sizeof(struct icmphdr), &orig_packet->iph,
	       orig_packet->iph.ihl * 4 + 8);

	packet->u.icmph.checksum = ~csum_partial(&packet->u,
						 ntohs(packet->iph.tot_len) - sizeof(struct iphdr),
						 0);
}

/* SRCIP TYPE CODE */
int main(int argc, const char *argv[])
{
	int rval;
	struct packet orig_packet, packet;
		
	if (argc < 4) {
		fprintf(stderr, "gen_err: Need at least SRCIP TYPE CODE\n");
		exit(1);
	}

	rval = read(STDIN_FILENO, &orig_packet, sizeof(orig_packet));
	if (rval < 0) {
		perror("gen_err: bad read");
		exit(1);
	}

	create_icmp(&packet, &orig_packet, argv[1], argv[2], argv[3]);

#if defined(__GLIBC__) && __GLIBC__ == 2
	packet.ehdr.ether_type = htons(ETHERTYPE_IP);
#else
	packet.ehdr.h_proto = htons(ETH_P_IP);
#endif
	write(STDOUT_FILENO, &packet, 16 + ntohs(packet.iph.tot_len));
	return 0;
}
