/* Simple program to generate and send ip packet.  */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <netdb.h>
#include <netinet/ip.h>
#if defined(__GLIBC__) && __GLIBC__ == 2
#include <netinet/udp.h>
#include <netinet/tcp.h>
#include <netinet/ip_icmp.h>
#include <net/ethernet.h>
#else
#include <linux/tcp.h>
#include <linux/udp.h>
#include <linux/icmp.h>
#include <linux/if_ether.h>
#endif
#include <sys/types.h>

static u_int16_t
csum_partial(void *buffer, unsigned int len, u_int16_t prevsum)
{
	u_int32_t sum = 0;
	u_int16_t *ptr = buffer;

	while (len > 1)  {
		sum += *ptr++;
		len -= 2;
	}
	if (len) {
		union {
			u_int8_t byte;
			u_int16_t wyde;
		} odd;
		odd.wyde = 0;
		odd.byte = *((u_int8_t *)ptr);
		sum += odd.wyde;
	}
	sum = (sum >> 16) + (sum & 0xFFFF);
	sum += prevsum;
	return (sum + (sum >> 16));
}

struct packet {
	u_int16_t pad;
#if defined(__GLIBC__) && __GLIBC__ == 2
	struct ether_header ehdr;
#else
	struct ethhdr ehdr;
#endif
	struct iphdr iph;
	union {
		struct tcphdr tcph;
		struct udphdr udph;
		struct icmphdr icmph;
		unsigned char raw[0];
	} u;
	unsigned char data[65535];
};

/* A few hardcoded protocols for 'all' and in case the user has no
   /etc/protocols */
struct pprot {
	char *name;
	unsigned short num;
};

static const struct pprot chain_protos[] = {
	{ "all", 0 },
	{ "tcp", IPPROTO_TCP },
	{ "udp", IPPROTO_UDP },
	{ "icmp", IPPROTO_ICMP },
	{0}
};

static int
string_to_number(const char *s, int min, int max)
{
	int number;
	char *end;

	number = (int)strtol(s, &end, 10);
	if (*end == '\0' && end != s) {
		/* we parsed a number, let's see if we want this */
		if (min <= number && number <= max)
			return number;
		else
			return -1;
	} else
		return -1;
}

static u_int16_t
parse_protocol(const char *s)
{
	struct protoent *pent;
	const struct pprot *pp;

	int proto = string_to_number(s, 0, 65535);

	if (proto != -1) return (unsigned short)proto;

	pent = getprotobyname(s);
	if (pent != NULL)
		return pent->p_proto;

	for (pp = &chain_protos[0]; pp->name; pp++)
		if (!strcasecmp(s, pp->name))
			return pp->num;

	fprintf(stderr, "gen_ip: unknown protocol `%s' specified\n", s);
	exit(1);
}

static struct in_addr *
dotted_to_addr(const char *dotted)
{
	static struct in_addr addr;
	unsigned char *addrp;
	char *p, *q;
	int onebyte, i;
	char buf[20];

	/* copy dotted string, because we need to modify it */
	strncpy(buf, dotted, sizeof(buf) - 1);
	addrp = (unsigned char *) &(addr.s_addr);

	p = buf;
	for (i = 0; i < 3; i++) {
		if ((q = strchr(p, '.')) == NULL)
			return (struct in_addr *) NULL;
		else {
			*q = '\0';
			if ((onebyte = string_to_number(p, 0, 255)) == -1)
				return (struct in_addr *) NULL;
			else
				addrp[i] = (unsigned char) onebyte;
		}
		p = q + 1;
	}

	/* we've checked 3 bytes, now we check the last one */
	if ((onebyte = string_to_number(p, 0, 255)) == -1)
		return (struct in_addr *) NULL;
	else
		addrp[3] = (unsigned char) onebyte;

	return &addr;
}

static void parse_header(struct packet *packet,
			 const char *srcip, const char *dstip,
			 const char *lenstr, const char *protocol,
			 u_int16_t fragoff,
			 int tos)
{
	u_int16_t proto = parse_protocol(protocol);
	int len = string_to_number(lenstr, sizeof(struct iphdr), 65535);
	u_int32_t saddr, daddr;
	const struct in_addr *addr;

	if (len < 0) {
		fprintf(stderr, "gen_ip: bad length `%s' specified\n", lenstr);
		exit(1);
	}

	addr = dotted_to_addr(srcip);
	if (!addr) {
		fprintf(stderr, "Bad src address `%s'\n", srcip);
		exit(1);
	}
	saddr = addr->s_addr;

	addr = dotted_to_addr(dstip);
	if (!addr) {
		fprintf(stderr, "Bad dst address `%s'\n", dstip);
		exit(1);
	}
	daddr = addr->s_addr;

	packet->iph = ((struct iphdr)
		      {
#if __BYTE_ORDER == __LITTLE_ENDIAN
			      5, 4,
#else
			      4, 5,
#endif
			      tos, htons(len), 0, fragoff, 255, proto, 0,
			      saddr, daddr });
	packet->iph.check = ~csum_partial(&packet->iph,
					  sizeof(packet->iph), 0);
}

/* TYPE CODE [ID SEQUENCE] */
static void
parse_icmp(struct packet *packet, char *args[])
{
	int type, code;

	if (!args[0] || (type = string_to_number(args[0], 0, 255)) < 0) {
		fprintf(stderr, "Bad icmp type: `%s'\n", args[0]);
		exit(1);
	}

	if (!args[1] || (code = string_to_number(args[1], 0, 255)) < 0) {
		fprintf(stderr, "Bad icmp code: `%s'\n", args[1]);
		exit(1);
	}

	packet->u.icmph.type = type;
	packet->u.icmph.code = code;
	packet->u.icmph.checksum = 0;

	if (type == 0 || type == 8) {
		int id, seq;

		if (!args[2] || (id = string_to_number(args[2], 0, 255)) < 0) {
			fprintf(stderr, "Bad icmp id: `%s'\n", args[2]);
			exit(1);
		}

		if (!args[3] || (seq = string_to_number(args[3], 0, 255)) < 0) {
			fprintf(stderr, "Bad icmp seq: `%s'\n", args[3]);
			exit(1);
		}
		packet->u.icmph.un.echo.id = htons(id);
		packet->u.icmph.un.echo.sequence = htons(seq);
	}

	packet->u.icmph.checksum = ~csum_partial(&packet->u,
						 ntohs(packet->iph.tot_len) - sizeof(struct iphdr),
						 0);
}

/* SRCPT DSTPT */
static void
parse_udp(struct packet *packet, char *args[])
{
	int spt, dpt;
	u_int16_t udplen = ntohs(packet->iph.tot_len) - sizeof(struct iphdr);
	struct {
		u_int32_t srcip, dstip;
		u_int8_t mbz, protocol;
		u_int16_t proto_len;
	} pseudo_header = { packet->iph.saddr, packet->iph.daddr, 0, IPPROTO_UDP, htons(udplen) };

	if (!args[0] || (spt = string_to_number(args[0], 0, 65535)) < 0) {
		fprintf(stderr, "Bad udp spt: `%s'\n", args[0]);
		exit(1);
	}

	if (!args[1] || (dpt = string_to_number(args[1], 0, 65535)) < 0) {
		fprintf(stderr, "Bad udp dpt: `%s'\n", args[1]);
		exit(1);
	}

	packet->u.udph = (struct udphdr){ htons(spt), htons(dpt), htons(udplen), 0 };

	packet->u.udph.check = ~csum_partial(&packet->u.udph, udplen,
					     csum_partial(&pseudo_header, sizeof(pseudo_header), 0));
}

static void
parse_flags(const char *string, struct tcphdr *tcph)
{
	memset(tcph, 0, sizeof(*tcph));

	if (strcmp(string, "SYN") == 0)
		tcph->syn = 1;
	else if (strcmp(string, "SYN/ACK") == 0)
		tcph->syn = tcph->ack = 1;
	else if (strcmp(string, "ACK") == 0)
		tcph->ack = 1;
	else if (strcmp(string, "RST") == 0)
		tcph->rst = 1;
	else if (strcmp(string, "FIN") == 0)
		tcph->fin = 1;
	else {
		fprintf(stderr, "Bad tcp flags `%s'\n", string);
		exit(1);
	}
}

static void
parse_tcp_options(char *opts, struct tcphdr *tcph)
{
	char *tok;
	u_int8_t options[15 * 4 - sizeof(struct tcphdr)];
	size_t i;

	/* Options are simple comma-separated number list. */
	for (i = 0, tok=strtok(opts, ","); tok; tok=strtok(NULL, ","), i++)
		options[i] = atol(tok);

	if (i >= sizeof(options)) {
		fprintf(stderr, "Too many TCP options `%s'\n", opts);
		exit(1);
	}
	if (i % 4 != 0) {
		fprintf(stderr, "TCP options must pad to 4 bytes\n");
		exit(1);
	}
	tcph->doff = (sizeof(struct tcphdr) + i) / 4;
	memcpy(tcph+1, options, i);
}

/* SRCPT DSTPT FLAGS [OPT=] [DATA] */
static void
parse_tcp(struct packet *packet, char *args[])
{
	int spt, dpt;
	u_int16_t tcplen = ntohs(packet->iph.tot_len) - sizeof(struct iphdr);
	struct {
		u_int32_t srcip, dstip;
		u_int8_t mbz, protocol;
		u_int16_t proto_len;
	} pseudo_header = { packet->iph.saddr, packet->iph.daddr, 0, IPPROTO_TCP, htons(tcplen) };

	packet->u.tcph.seq = packet->u.tcph.ack_seq = 0;

	if (!args[0] || (spt = string_to_number(args[0], 0, 65535)) < 0) {
		fprintf(stderr, "Bad tcp spt: `%s'\n", args[0]);
		exit(1);
	}

	if (!args[1] || (dpt = string_to_number(args[1], 0, 65535)) < 0) {
		fprintf(stderr, "Bad tcp dpt: `%s'\n", args[1]);
		exit(1);
	}

	parse_flags(args[2], &packet->u.tcph);

	if (args[3] && strncmp(args[3], "OPT=", 4) == 0) {
		parse_tcp_options(args[3] + 4, &packet->u.tcph);
		args++;
	} else {
		packet->u.tcph.doff = sizeof(struct tcphdr) / 4;
	}

	if (args[3] && strncmp(args[3], "SYN=", 4) == 0) {
		/* SYN number. */
		packet->u.tcph.seq = htonl(atol(args[3]+4));
		args++;
	}

	if (args[3] && strncmp(args[3], "ACK=", 4) == 0) {
		/* ACK number. */
		packet->u.tcph.ack_seq = htonl(atol(args[3]+4));
		args++;
	}

	if (args[3] && strcmp(args[3], "DATA") == 0) {
		/* Read DATA for packet from stdin. */
		if (read(STDIN_FILENO,
			 (void *)&packet->u.tcph + packet->u.tcph.doff * 4,
			 tcplen - (packet->u.tcph.doff * 4)) < 0) {
			perror("Reading TCP DATA");
			exit(1);
		}
	}

	packet->u.tcph.source = htons(spt);
	packet->u.tcph.dest = htons(dpt);
	packet->u.tcph.window = packet->u.tcph.check = packet->u.tcph.urg_ptr
		= 0;

	packet->u.tcph.check = ~csum_partial(&packet->u.tcph, tcplen,
					     csum_partial(&pseudo_header,
							  sizeof(pseudo_header), 0));
}

static int
parse_mac(const char *str, u_int8_t *mac)
{
	unsigned int i = 0;

	if (strlen(str) != ETH_ALEN*3-1)
		return 0;

	for (i = 0; i < ETH_ALEN; i++, mac++) {
		long number;
		char *end;

		number = strtol(str + i*3, &end, 16);

		if (end == str + i*3 + 2
		    && number >= 0
		    && number <= 255)
			*mac = (u_int8_t)number;
		else
			return 0;
	}
	return 1;
}

/* SRCIP DSTIP LEN PROTOCOL */
int main(int argc, char *argv[])
{
	struct packet packet;
	int tos = 0;
	u_int16_t fragoff = 0;

	if (argc < 5) {
		fprintf(stderr, "gen_ip: Need at least [FRAG=START] [MF] [MAC=<hex>] [TOS=n] SRCIP DSTIP LEN PROTOCOL\n");
		exit(1);
	}

	if (argc > 5 && strcmp(argv[1], "MF") == 0) {
		fragoff |= IP_MF;
		argc--;
		argv++;
	}
	if (argc > 5 && strncmp(argv[1], "FRAG=", 5) == 0) {
		int frag = atoi(argv[1]+5);

		if (frag < 1 || frag > 65528 || (frag % 8) != 0) {
			fprintf(stderr, "`%s' not >= 8, <= 65528, mod 8 = 0\n",
				argv[1]+5);
			exit(1);
		}
		fragoff |= ((unsigned)frag / 8);
		argc--;
		argv++;
	}

	if (argc > 5 && strncmp(argv[1], "MAC=", 4) == 0) {
		if (!parse_mac(argv[1]+4,
#if defined(__GLIBC__) && __GLIBC__ == 2
			       packet.ehdr.ether_shost
#else
			       packet.ehdr.h_source
#endif
			)) {
			fprintf(stderr, "Invalid mac address.\n");
			exit(1);
		}
		argc--;
		argv++;
	}

	if (argc > 5 && strncmp(argv[1], "TOS=", 4) == 0) {
		tos = atoi(argv[1]+4);

		if (tos == 0 || tos > 255) {
			fprintf(stderr, "`%s' not >= 0, <= 255\n",
				argv[1]+4);
			exit(1);
		}
		argc--;
		argv++;
	}
			

	parse_header(&packet, argv[1], argv[2], argv[3], argv[4],
		     htons(fragoff), tos);
	switch (packet.iph.protocol) {
	case IPPROTO_ICMP:
		parse_icmp(&packet, argv + 5);
		break;
	case IPPROTO_TCP:
		parse_tcp(&packet, argv + 5);
		break;
	case IPPROTO_UDP:
		parse_udp(&packet, argv + 5);
		break;
	default:
		if (argc != 5) {
			fprintf(stderr, "Extra args for unknown protocol\n");
			exit(1);
		}
	}

#if defined(__GLIBC__) && __GLIBC__ == 2
	packet.ehdr.ether_type = htons(ETHERTYPE_IP);
#else
	packet.ehdr.h_proto = htons(ETH_P_IP);
#endif
	write(STDOUT_FILENO, &packet, 16 + ntohs(packet.iph.tot_len));
	return 0;
}
