/* Intercepts packet from given netfilter hook, and outputs to stdout:
   Useful for testing with rcv_ip. */
#include <sys/types.h>
#include <net/if.h>

#include "device/netfilter_dev.h"
#include <linux/netfilter_ipv4.h>

#include <fcntl.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>

#define NETFILTER_DEVICE_NAME 	"/dev/netfilter_ipv4"

int main(int argc, const char *argv[])
{
	int fd;
	struct nfdev_verdict verdict;
	struct nfdev_condition cond = {0xFFFFFFFF, 0, 0};
	unsigned int user_verdict, count, timeout, wait, mangle, i;
	struct {
		struct nfdev_head head;
		unsigned char packet[65536];
	} pack;

	if (argc < 5 || argc > 7) {
		fprintf(stderr,
			"Usage: %s hook-name verdict timeout count"
			" [wait] [`mangle']\n",
			argv[0]);
		return 1;
	}

	fd = open(NETFILTER_DEVICE_NAME, O_RDWR);
	if (fd < 0) {
		fprintf(stderr, "opening %s: %s\n",
			NETFILTER_DEVICE_NAME, strerror(errno));
		return 1;
	}

	/* Starts with "catch-all" condition. */
	if (ioctl(fd, NFDIOSDELCOND, &cond) < 0) {
		perror("deleting original condition");
		return 1;
	}

	/* FIXME: blow off any captured packets. --RR */
	if (strcasecmp(argv[1], "LOCAL_IN") == 0)
		cond.hookmask = 1 << NF_IP_LOCAL_IN;
	else if (strcasecmp(argv[1], "PRE_ROUTING") == 0)
		cond.hookmask = 1 << NF_IP_PRE_ROUTING;
	else if (strcasecmp(argv[1], "FORWARD") == 0)
		cond.hookmask = 1 << NF_IP_FORWARD;
	else if (strcasecmp(argv[1], "POST_ROUTING") == 0)
		cond.hookmask = 1 << NF_IP_POST_ROUTING;
	else if (strcasecmp(argv[1], "LOCAL_OUT") == 0)
		cond.hookmask = 1 << NF_IP_LOCAL_OUT;
	else {
		fprintf(stderr, "Unknown hook `%s'\n", argv[1]);
		return 1;
	}

	if (ioctl(fd, NFDIOSADDCOND, &cond) < 0) {
		perror("adding new condition");
		return 1;
	}

	if (strcasecmp(argv[2], "ACCEPT") == 0)
		user_verdict = NF_ACCEPT;
	else if (strcasecmp(argv[2], "DROP") == 0)
		user_verdict = NF_DROP;
	else {
		fprintf(stderr, "Illegal verdict `%s'\n", argv[2]);
		return 1;
	}

	timeout = atoi(argv[3]);
	if (timeout < 1) {
		fprintf(stderr, "Illegal timeout `%s'\n", argv[3]);
		return 1;
	}

	count = atoi(argv[4]);
	if (count < 1) {
		fprintf(stderr, "Illegal count `%s'\n", argv[4]);
		return 1;
	}

	if (argc > 5) {
		wait = atoi(argv[5]);
		if (wait < 1) {
			fprintf(stderr, "Illegal wait `%s'\n", argv[5]);
			return 1;
		}
		argc--; argv++;
	} else
		wait = 0;

	if (argc > 5 && strcmp(argv[5], "mangle") == 0) {
		mangle = 1;
		argc--; argv++;
	} else
		mangle = 0;

	if (argc > 5) {
		fprintf(stderr, "Error in args: too many.\n");
		exit(1);
	}

	for (i = 0; i < count; i++) {
		int rval;

		/* We'll die if we receive this. OK. */
		alarm(timeout);

		rval = read(fd, &pack, sizeof(pack));
		if (rval < sizeof(pack.head)) {
			perror("short read");
			return 1;
		}

		if (getenv("VERBOSE"))
			fprintf(stderr,
				"in=%s, out=%s, mk=%lu, len=%u v=%s\n",
				pack.head.iniface, pack.head.outiface,
				pack.head.mark, pack.head.pktlen,
				user_verdict == NF_ACCEPT ? "ACCEPT"
				: (user_verdict == NF_DROP ? "DROP" : "?"));

		/* Bogus ethernet header for rcv_ip */
		write(STDOUT_FILENO, pack.packet - 16,
		      16 + rval - sizeof(pack.head));

		/* This usually wipes out alarm, but that's OK since
                   the rest shouldn't block. */
		if (wait)
			sleep(wait);

		if (mangle) {
			/* Read in new packet from stdin. */

			/* Discard header. */
			read(STDIN_FILENO, pack.packet, 16);
			rval = read(STDIN_FILENO, pack.packet,
				    sizeof(pack.packet));
			if (rval < 0) {
				perror("Reading new packet");
				return 1;
			}
			pack.head.pktlen = rval;

			if (lseek(fd, sizeof(pack.head), SEEK_SET)
			    == (off_t)-1) {
				perror("Seeking");
				return 1;
			}
			if (write(fd, pack.packet, rval) != rval) {
				perror("Writing new packet");
				return 1;
			}
			fprintf(stderr, "New packet written\n");
		}

		verdict = ((struct nfdev_verdict)
		   { user_verdict, pack.head.pktlen, pack.head.mark,
		     pack.head.reason });
		if (ioctl(fd, NFDIOSVERDICT, &verdict) != 0) {
			perror("ioctl");
			return 1;
		}

		if (getenv("VERBOSE") && wait)
			fprintf(stderr, "intercept: reinjected.\n");

	}
	return 0;
}

	

