/* Simple program to receive and print IP packet.  */
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <netinet/ip.h>
#if defined(__GLIBC__) && __GLIBC__ == 2
#include <net/ethernet.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <netinet/ip_icmp.h>
#else
#include <linux/if_ether.h>
#include <linux/tcp.h>
#include <linux/udp.h>
#include <linux/icmp.h>
#endif
#include <sys/types.h>

#define IP_PARTS_NATIVE(n)			\
(unsigned int)((n)>>24)&0xFF,			\
(unsigned int)((n)>>16)&0xFF,			\
(unsigned int)((n)>>8)&0xFF,			\
(unsigned int)((n)&0xFF)

#define IP_PARTS(n) IP_PARTS_NATIVE(ntohl(n))

#ifndef IP_OFFSET
#define	IP_OFFSET 0x1FFF
#endif

static u_int16_t
csum_partial(void *buffer, unsigned int len, u_int16_t prevsum)
{
	u_int32_t sum = 0;
	u_int16_t *ptr = buffer;

	while (len > 1)  {
		sum += *ptr++;
		len -= 2;
	}
	if (len) {
		union {
			u_int8_t byte;
			u_int16_t wyde;
		} odd;
		odd.wyde = 0;
		odd.byte = *((u_int8_t *)ptr);
		sum += odd.wyde;
	}
	sum = (sum >> 16) + (sum & 0xFFFF);
	sum += prevsum;
	return (sum + (sum >> 16));
}

struct packet {
	u_int16_t pad;
#if defined(__GLIBC__) && __GLIBC__ == 2
	struct ether_header ehdr;
#else
	struct ethhdr ehdr;
#endif
	struct iphdr iph;
	union {
		struct tcphdr tcph;
		struct udphdr udph;
		struct icmphdr icmph;
		unsigned char raw[0];
	} u;
	unsigned char data[65535];
};

static int
string_to_number(const char *s, int min, int max)
{
	int number;
	char *end;

	number = (int)strtol(s, &end, 10);
	if (*end == '\0' && end != s) {
		/* we parsed a number, let's see if we want this */
		if (min <= number && number <= max)
			return number;
		else
			return -1;
	} else
		return -1;
}

int main(int argc, const char *argv[])
{
	struct packet packet;
	int wait_time, iterations, print_data;
	int rval;
	int bad_pkt = 0;
	u_int16_t plen;

	if (argc != 3 && argc != 4) {
		fprintf(stderr, "rcv_ip: Need WAIT-TIME ITERATIONS [DATA]\n");
		exit(1);
	}

	if ((wait_time = string_to_number(argv[1], 1, 3600)) < 0) {
		fprintf(stderr, "rcv_ip: bad wait-time `%s'\n", argv[1]);
		exit(1);
	}

	if ((iterations = string_to_number(argv[2], 1, 100000)) < 0) {
		fprintf(stderr, "rcv_ip: bad iterations `%s'\n", argv[2]);
		exit(1);
	}

	if (argv[3]) {
		if (strcmp(argv[3], "DATA") != 0) {
			fprintf(stderr, "rcv_ip: bad arg `%s'\n", argv[3]);
			exit(1);
		}
		print_data = 1;
	} else {
		print_data = 0;
	}		

	for (; iterations; iterations--) {
		/* We'll die if we receive this. OK. */
		alarm(wait_time);
		rval = read(STDIN_FILENO, &packet, sizeof(packet));
		if (rval < 0) {
			perror("rcv_ip: bad read");
			exit(1);
		}

		if (rval < 16 + sizeof(struct iphdr)) {
			printf("-TRUNCATED-\n");
			bad_pkt = 1;
			continue;
		}

		if (csum_partial(&packet.iph, sizeof(packet.iph), 0)
		    != 0xFFFF) {
			printf("-BAD IP CSUM- ");
			bad_pkt = 1;
		}

		/* Output in a form suitable for gen_ip */
		if (ntohs(packet.iph.frag_off) & IP_OFFSET)
			printf("FRAG=%u ", ntohs(packet.iph.frag_off)*8);
		if (ntohs(packet.iph.frag_off) & IP_MF)
			printf("MF ");

		printf("%u.%u.%u.%u %u.%u.%u.%u %u %u ",
		       IP_PARTS(packet.iph.saddr), IP_PARTS(packet.iph.daddr),
		       ntohs(packet.iph.tot_len), packet.iph.protocol);

		plen = ntohs(packet.iph.tot_len) - sizeof(struct iphdr);

		rval -= 16 + sizeof(struct iphdr);

		/* Don't print per-proto for non-first fragments */
		if (ntohs(packet.iph.frag_off) & 0x1fff)
			goto end;

		/* FIXME: Doesn't handle IP options. */
		switch (packet.iph.protocol) {
		case IPPROTO_ICMP:
			if (rval < sizeof(struct icmphdr)) {
				printf("-TRUNCATED- ");
				bad_pkt = 1;
				break;
			}

			printf("%u %u ",
			       packet.u.icmph.type, packet.u.icmph.code);
			if (packet.u.icmph.type == 0
			    || packet.u.icmph.type == 8) {
				printf("%u %u ",
				       ntohs(packet.u.icmph.un.echo.id),
				       ntohs(packet.u.icmph.un.echo.sequence));
			}

			if (rval < plen) {
				printf("-TRUNCATED- ");
				bad_pkt = 1;
				break;
			}

			if (!(ntohs(packet.iph.frag_off) & IP_MF)
			    && csum_partial(&packet.u, plen, 0) != 0xFFFF) {
				printf("-BAD ICMP CSUM- ");
				bad_pkt = 1;
			}
			break;

		case IPPROTO_UDP: {
			struct {
				u_int32_t srcip, dstip;
				u_int8_t mbz, protocol;
				u_int16_t proto_len;
			} pseudo_header =
			  { packet.iph.saddr, packet.iph.daddr,
			    0, IPPROTO_UDP, htons(plen) };

			if (rval < sizeof(struct udphdr)) {
				printf("-TRUNCATED- ");
				bad_pkt = 1;
				break;
			}

			printf("%u %u ",
			       ntohs(packet.u.udph.source),
			       ntohs(packet.u.udph.dest));

			if (rval < plen) {
				printf("-TRUNCATED- ");
				bad_pkt = 1;
				break;
			}

			if (!(ntohs(packet.iph.frag_off) & IP_MF)
			    && csum_partial(&packet.u, plen,
					    csum_partial(&pseudo_header,
							 sizeof(pseudo_header),
							 0)) != 0xFFFF) {
				printf("-BAD UDP CSUM- ");
				bad_pkt = 1;
			}
			break;
		}
		
		case IPPROTO_TCP: {
			unsigned int i;
			struct {
				u_int32_t srcip, dstip;
				u_int8_t mbz, protocol;
				u_int16_t proto_len;
			} pseudo_header =
			  { packet.iph.saddr, packet.iph.daddr,
			    0, IPPROTO_TCP, htons(plen) };

			if (rval < sizeof(struct tcphdr)) {
				printf("-TRUNCATED- ");
				bad_pkt = 1;
				break;
			}

			printf("%u %u ",
			       ntohs(packet.u.tcph.source),
			       ntohs(packet.u.tcph.dest));
			if (packet.u.tcph.rst)
				printf("RST ");
			else if (packet.u.tcph.fin)
				printf("FIN ");
			else if (packet.u.tcph.syn && packet.u.tcph.ack)
				printf("SYN/ACK ");
			else if (packet.u.tcph.syn)
				printf("SYN ");
			else if (packet.u.tcph.ack)
				printf("ACK ");

			if (rval < plen) {
				printf("-TRUNCATED- ");
				bad_pkt = 1;
				break;
			}

			if (packet.u.tcph.doff*4 != sizeof(struct tcphdr)) {
				char sep = '=';
				printf("OPT");
				for (i = sizeof(struct tcphdr);
				     i < packet.u.tcph.doff * 4;
				     i++) {
					printf("%c%u",
					       sep,
					       i[(u_int8_t *)&packet.u.tcph]);
					sep = ',';
				}
				printf(" ");
			}
			if (packet.u.tcph.seq)
				printf("SYN=%u ", ntohl(packet.u.tcph.seq));
			if (packet.u.tcph.ack_seq)
				printf("ACK=%u ",
				       ntohl(packet.u.tcph.ack_seq));
			if (!(ntohs(packet.iph.frag_off) & IP_MF)
			    && csum_partial(&packet.u, plen,
					    csum_partial(&pseudo_header,
							 sizeof(pseudo_header),
							 0)) != 0xFFFF) {
				printf("-BAD TCP CSUM- ");
				bad_pkt = 1;
			}
			if (print_data) {
				printf("DATA\n");
				fflush(stdout);
				write(STDOUT_FILENO,
				      (void *)&packet.u.tcph
				      + packet.u.tcph.doff * 4,
				      plen - packet.u.tcph.doff * 4);
			}
			break;
		}
		}
	end:
		printf("\n");
	}

	return bad_pkt ? 2 : 0;
}
